/*
 * ported to Linux from Francis Dupont (INRIA Rocquencourt)'s IPv6 
 * implementation on Net/FreeBSD.
 *
 * Last updates:
 *
 * 08/09/97: bbrodard@sophia.inria.fr
 *
 * TODO:
 *   - Interaction w/ sys. adm.
 *   - Sit and Loopback configuration.
 *   - Check IPV6_MULTICAST_LOOP option.
 *   - Differentiate tentative and valid address in kernel (flag)
 *   - Media other than IEEE802.
 */
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <sysexits.h>
#include <string.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/queue.h>
#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in_systm.h>
#include <linux/netdevice.h>
#include <linux/in.h>
#include <linux/dret/ipv6.h>
#include <linux/dret/icmpv6.h>
#include <linux/socket.h>
#include <linux/string.h>
#include <linux/sockios.h>

#include <asm/atomic.h>
#include <asm/checksum.h>
#include <asm/types.h>

#include <net/dret/addrconf.h>
#include <net/dret/nd.h>
#include <net/route.h>
#include <net/if.h>
/* FIXME */
#include "/0/rodeo/bbrodard/CVS_MINE/IPv6_DRET/BIND/include/arpa/inet.h"


#define MAX_INITIALIZATION_DELAY	3   /* seconds */
#define DUPL_ADDR_RETRANS_TIMER		3   /* seconds */
#define MAX_DUPL_ADDR_RETRANS		2

#define Malloc(p, t)	(p = (struct t *)malloc(sizeof(struct t)))

struct interface {
   LIST_ENTRY(interface) if_list;	/* list of interfaces */
   
   unsigned char laddr6[MAX_ADDR_LEN];   /* link address */
   unsigned char addr_len;      /* link address length */
   struct sockaddr_in6 *addr6;	/* IPv6 address */
   struct sockaddr_in6 *mask6;	/* IPv6 netmask */
   struct sockaddr_in *addr4;	/* IPv4 address */
   struct sockaddr_in *mask4;	/* IPv4 netmask */
   u_int flags;			/* copy of kernel flags */
   u_int index;			/* copy of kernel intf index */
   u_int kind;			/* kind */
   u_int unit;			/* rank in kind */
   char name[IFNAMSIZ];		/* name */
};
LIST_HEAD(, interface) interfaces;
struct interface *main_if;

#define IFK_IEEE	1		/* Ethernet, FDDI, ... */
#define IFK_LOOP	2		/* Loopback */
#define IFK_SIT		3		/* IPv6 over IPv4 */
#define IFK_LLINK	4		/* pseudo-IEEE */
#define IFK_P2P		5		/* point to point */
#define IFK_NBMA	6		/* NBMA */
#define IFK_IPV4	7		/* IPv4 only */

int ifk_ieee  = 0;
int ifk_loop  = 0;
int ifk_sit   = 0;
int ifk_llink = 0;
int ifk_p2p   = 0;
int ifk_nbma  = 0;
int ifk_ipv4  = 0;

#define MAXPACKET	576	        /* max packet size */
u_char packet[MAXPACKET];

int pid;

struct ifreq ifr;
struct sockaddr_in6 dst = {AF_INET6 };

char usage[] = "Usage: autoconf6 [-v] [interface]\n";

int s, r, u, v;			/* Socket file descriptors */

/* Common variables */
int verbose = 0;		/* extra printfs */
int install_sit = 1;		/* install SIT interface */
int dring = 0;			/* timeout expired */
double delay;			/* initial timeout in seconds */
u_int reload;			/* retrans timeout in seconds */
int ntosend = 0;		/* packet to send counter */
int ninflow = 0;		/* packet on the cable counter */
char *mname = NULL;		/* main interface name */
char inet_ntop_ret[INET6_ADDRSTRLEN]; /* to be passed to */

void timer();
void send_ndsol(struct interface *);
void recv_pack(char *, int);
int  in_cksum(u_short *, int);
void init();
struct interface *find_main(char *);
void setup_llink();
void setup_ieee(struct interface *);
void setup_sit(struct interface *);
void setup_loop(struct interface *);
void setup_ieee2(struct interface *);
void setup_loop2(struct interface *);
void booting(struct interface *, int);
void add_default_route(struct interface *, u_int);
void collision(struct interface *);

void 
main(int argc, char *argv[]) {
   register struct interface *ifp;
   struct sigaction sa;
   char **av = argv;
   u_int val;

   pid = getpid();
   
   /* 
    * Read arguments
    */
   argc--, av++;
   while (argc > 0 && *av[0] == '-') {
      while (*++av[0]) 
	 switch (*av[0]) {
	 case 'v':
	    verbose++;
	    break;
	 case '6':
	    install_sit = 0;
	    break;
	 default:
	    (void) fprintf(stderr, usage);
	    exit(EX_USAGE);
	 }
      argc--, av++;
   }

   if (argc > 0) {
      mname = *av;
      argc--;
   }
   
   if (argc != 0) {
      (void) fprintf(stderr, "Extra paramaters\n");
      (void) fprintf(stderr, usage);
      exit(EX_USAGE);
   }
   
   srandom((int)gethostid());
   setlinebuf(stdout);
   
   /* 
    * force EINTR on ICMP reads 
    */
   sa.sa_handler = timer;
   sa.sa_mask = 0;
   sa.sa_flags = SA_INTERRUPT;
   
   (void) sigaction(SIGALRM, &sa, (struct sigaction *)NULL);
   
   /* 
    * Open ICMPv6 "s" socket (send/receive raw packets) 
    */
   if ((s = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
      perror("socket (ICMP)");
      exit(EX_CANTCREAT);
   }

   /* Header included on outgoing packets */
   val = 1;
   if (setsockopt(s, SOL_IPV6, IPV6_HDRINCL, &val, sizeof(val)) < 0) {
      perror("setsockopt (IP_HDRINCL)");
      exit(EX_OSERR);
   }
	
   /* Anti spoofing hop number: 255 */    
   val = ICMP6_ND_HOPS;
   if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
		  &val, sizeof(val)) < 0) {
      perror("setsockopt (IPV6_MULTICAST_HOPS)");
      exit(EX_OSERR);
   }
	
   /*
     val = 1;
     if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
     &val, sizeof(val)) < 0) {
     perror("setsockopt (IPV6_MULTICAST_LOOP)");
     exit(EX_OSERR);
     }
     */

   /* 
    * Open UDP/IPv6 "u" socket (ioctl) 
    */
   if ((u = socket(PF_INET6, SOCK_DGRAM, 0)) < 0) {
      perror("socket (UDP6)");
      exit(EX_CANTCREAT);
   }
   
   /* 
    * Open UDP/IPv4 "v" socket (ioctl)
    */
   if ((v = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
      perror("socket (UDP4)");
      exit(EX_CANTCREAT);
   }

   /* 
    * Get interfaces 
    */
   init();

   /*
    * Find the main IEEE LAN interface and put it first with unit 0.
    */
   main_if = find_main(mname);   
   if (verbose)
      printf("default IEEE interface is %s\n", main_if->name);

   /* Main interface unit is 0 */
   if (main_if->unit != 0) {
      for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
	 if ((ifp->kind == IFK_IEEE) && (ifp->unit == 0))
	    break;
      ifp->unit = main_if->unit;
      main_if->unit = 0;
   }
   
   /* Main interface is head of list */
   if (interfaces.lh_first != main_if) {
      LIST_REMOVE(main_if, if_list);
      LIST_INSERT_HEAD(&interfaces, main_if, if_list);
   }

   /*
    * Put link-local address on LAN IEEE interfaces.
    */
   for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
      if (ifp->kind == IFK_IEEE)
	 setup_ieee(ifp);
			
   /*
    * Setup other interfaces.
    */
   
   
   /*
    * add a direct default route on the main interface
    * (every thing is on-link).
    */
   add_default_route(main_if, 0);
   exit(EX_OK);
}

/*
 * Note that timeout has rung.
 */
void
timer()
{
   dring++;
}

/*
 * Send an ICMPv6 Neighbor Solicitation packet on an interface.
 */
void
send_ndsol(struct interface * ifp) {
   static u_char outpack[MAXPACKET];
   register struct ipv6hdr *ip;
   register struct n_adv_sol *icp;
   int len, tlen, cc;
   unsigned int csum;
   struct in6_addr src = IN6ADDR_ANY_INIT;
   
   if (verbose) {
      printf("Sending ND sol for %s on %s\n",
	     inet_ntop(AF_INET6, &ifp->addr6->sin6_addr,
		       inet_ntop_ret, INET6_ADDRSTRLEN),
	     ifp->name);
   }

   /* 
    * Compute the solicited-node multicast address 
    */
   CLR_ADDR6(dst.sin6_addr);
   dst.sin6_addr.s6_addr8[0] = 0xff;
   dst.sin6_addr.s6_addr8[1] = 2;
   dst.sin6_addr.s6_addr8[11] = 1;
   dst.sin6_addr.s6_addr8[12] = ifp->addr6->sin6_addr.s6_addr8[12];
   dst.sin6_addr.s6_addr8[13] = ifp->addr6->sin6_addr.s6_addr8[13];
   dst.sin6_addr.s6_addr8[14] = ifp->addr6->sin6_addr.s6_addr8[14];
   dst.sin6_addr.s6_addr8[15] = ifp->addr6->sin6_addr.s6_addr8[15];

   len = sizeof(struct n_adv_sol);
   tlen = len + sizeof(struct ipv6hdr);
   bzero((caddr_t)outpack, tlen);

   ip = (struct ipv6hdr *)outpack;
   ip->next_hdr = IPPROTO_ICMPV6;
   COPY_ADDR6(ip->daddr, dst.sin6_addr);
   ip->payload = htons(len);
   
   icp = (struct n_adv_sol *)(ip + 1);
   icp->hdr.type = ICMP6_SOLICITATION_ND;
   COPY_ADDR6(icp->target, ifp->addr6->sin6_addr);
   
   /* other fields are 0 */

   /* Compute ICMPv6 checksum here */
   csum = csum_partial((char *)icp, len, 0);
   icp->hdr.csum = csum_ipv6_magic(&src, &(dst.sin6_addr), len, 
				   IPPROTO_ICMPV6, csum);
					 
   /* finish IPv6 header */
   ip->vrprflow = IPV6_VERSION | ICMP6_ND_PRIORITY;
   ip->hop_limit = ICMP6_ND_HOPS;
   
   /* FIXME */
   if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		  &ifp->index, sizeof(ifp->index)) < 0) {
      perror("setsockopt (IPV6_MULTICAST_IF)");
      return;
   }

   cc = sendto(s, (char *)outpack, tlen, 0,
	       (struct sockaddr *)&dst, sizeof (struct sockaddr_in6));

   if (cc < 0 || cc != tlen)  {
      fprintf(stderr, "wrote %d chars, ret=%d\n", tlen, cc);
      if (cc < 0) {
	 perror("sendto");
	 return;
      }
   }
   ninflow++;
}

/* Receive an ICMPv6 packet */
void
recv_pack(char *buf, int cc) {
   register struct ipv6hdr *ip;
   register struct n_adv_sol *icp;
   register struct interface *ifp;

   ip = (struct ipv6hdr *)buf;
   if (cc < sizeof(struct ipv6hdr) + sizeof(struct icmpv6_hdr)) {
      if (verbose)
	 printf("ICMPv6: too short (%d)\n", cc);
      return;
   }
   
   cc -= sizeof(*ip);
   icp = (struct n_adv_sol *)(ip + 1);

   switch (icp->hdr.type) {
   case ICMP6_ADVERTISMENT_ND:
      /* Basic checks */
      if (icp->hdr.code != 0) {
	 if (verbose)
	    printf("ICMPv6 ND Adv: code=%d\n",
		   icp->hdr.code);
	 return;
      }
      if ((unsigned)cc < sizeof(struct n_adv_sol)) {
	 if (verbose)
	    printf("ICMPv6 ND Adv: too short %d\n", cc);
	 return;
      }
      
      /* Look for interface */
      for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
	 if (((ifp->kind == IFK_IEEE) ||
	      (ifp->kind == IFK_LLINK)) &&
	     ifp->addr6 &&
	     SAME_ADDR6(icp->target, ifp->addr6->sin6_addr))
	    break;
      if (ifp == NULL)
	 return;
      
      /* Collision occured */
      if (verbose) {
	 fprintf(stderr, "Adv collision on %s", ifp->name);
	 fprintf(stderr, " src=%s",
		 inet_ntop(AF_INET6, &ip->saddr,
			   inet_ntop_ret, INET6_ADDRSTRLEN));
	 fprintf(stderr, " dst=%s",
		 inet_ntop(AF_INET6, &ip->daddr, inet_ntop_ret,
			   INET6_ADDRSTRLEN));
	 fprintf(stderr, " target=%s",
		 inet_ntop(AF_INET6, &icp->target, inet_ntop_ret,
			   INET6_ADDRSTRLEN));
      }
      collision(ifp);
      exit(EX_DATAERR);
      
   case ICMP6_SOLICITATION_ND:
      /* Basic checks */
      if (icp->hdr.code != 0) {
	 if (verbose)
	    printf("ICMPv6 ND Sol: code=%d\n",
		   icp->hdr.code);
	 return;
      }
      if ((unsigned)cc < sizeof(struct n_adv_sol)) {
	 if (verbose)
	    printf("ICMPv6 ND Sol: too short %d\n", cc);
	 return;
      }
      /* Look for interface */
      for (ifp = interfaces.lh_first; ifp;
	   ifp = ifp->if_list.le_next)
	 if (((ifp->kind == IFK_IEEE) ||
	      (ifp->kind == IFK_LLINK)) &&
	     ifp->addr6 &&
	     SAME_ADDR6(icp->target, ifp->addr6->sin6_addr)
	     )
	    break;
      if (ifp == NULL ||                   /* Not for us */
	  !IS_ANYADDR6(ip->saddr) ||       /* Not DAD */
	  --ninflow >= 0) {
	 if (verbose)
	    printf("got my ND sol\n");
	 return;
      }
      /* Collision occured */
      fprintf(stderr, "Sol collision on %s", ifp->name);
      fprintf(stderr, " src=%s",
	      inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
			INET6_ADDRSTRLEN));
      fprintf(stderr, " dst=%s",
	      inet_ntop(AF_INET6, &ip->daddr, inet_ntop_ret,
			INET6_ADDRSTRLEN));
      
      fprintf(stderr, " target=%s",
	      inet_ntop(AF_INET6, &icp->target, inet_ntop_ret,
			INET6_ADDRSTRLEN));
      
      collision(ifp);
      exit(EX_DATAERR);
      break;
   }
}

/*
 * Get interfaces with their addresses.
 */
void
init() {
   register struct interface *ifp;   /* Application interface list */
   register struct ifreq     *ifr;   /* Interface list returned by kernel */
   struct ifreq     ifr2;
   struct ifconf    ifc;

   int i;
   unsigned int tp;
   char buf[1024], buf2[1024];
   
   LIST_INIT(&interfaces);
   ifp = NULL;
   
   /*
    * Get all interfaces
    */
   ifc.ifc_len = sizeof(buf);
   ifc.ifc_buf = buf;
   if (ioctl(u, SIOCGIFCONF, &ifc) < 0) {
      fprintf(stderr, "SIOCGIFCONF: %s\n", strerror(errno));
      return;
   }
   
   /*
    * Scan interface list returned by kernel
    */
   ifr = ifc.ifc_req;
   printf("-%d-%d-\n", ifc.ifc_len / sizeof(struct ifreq), sizeof(struct ifreq));
   for (i = ifc.ifc_len / sizeof(struct ifreq); i-- > 0; ifr++) {
      fprintf(stderr, "Analyzing %s\n", ifr->ifr_name);
      if ((sscanf(ifr->ifr_name, "sit%u", &tp) == 1) ||
	  (strcmp(ifr->ifr_name, "lo") == 0) ||
	  (strncmp(ifr->ifr_name, "ct", 2) == 0)) {
	  /* No external autoconf for sit and lo devices*/
	 fprintf(stderr, "   No external autoconf for %s\n", ifr->ifr_name);
	 continue;
      }
      /*
       * New element in application interface list
       */
      if (Malloc(ifp, interface) == NULL) {
	 fprintf(stderr, "   Out of memory\n");
	 exit(EX_CANTCREAT);
      }
      bzero((caddr_t)ifp, sizeof(*ifp));
      
      /* index */
      if (ioctl(u, SIOGIFINDEX, ifr) < 0){
	 fprintf(stderr, "   Can't get %s 's index\n", ifr->ifr_name);
	 goto error;
      }
      ifp->index = ifr->ifr_ifindex;

      /* flags */
      if (ioctl(u, SIOCGIFFLAGS, ifr) < 0){
	 fprintf(stderr, "   Can't get %s 's flags\n", ifr->ifr_name);
	 goto error;
      }
      ifp->flags = ifr->ifr_flags;

      /*
       * IEEE802
       */
      if ((ifr->ifr_flags & IFF_BROADCAST) &&
	  (ifr->ifr_flags & IFF_MULTICAST)) {
	 if ((ifr->ifr_flags & IFF_UP)) {
	    /* link-layer address */
	    if (ioctl(v, SIOCGIFHWADDR, ifr) < 0){
	       fprintf(stderr, "   Can't get %s 's hardware address?\n",
		       ifr->ifr_name);
	       goto error;
	    }
	    memcpy(ifp->name, ifr->ifr_name, IFNAMSIZ);
	    memcpy(ifp->laddr6, ifr->ifr_hwaddr.sa_data, 8);
	    
	    ifp->addr_len = 6; /* FIXME? */
	    ifp->kind = IFK_IEEE;
	    ifp->unit = ifk_ieee++;
	 } else {
	    fprintf(stderr, "   Interface %s is not up\n", ifr->ifr_name);
	    goto error;
	 }
      }
      	 
      /* 
       * point to point
       */
      
      /*
       * nbma 
       */
	
      if (ifp->kind != 0) {
	 if (verbose)
	    printf("   Got interface %s\n", ifp->name);
      } else {
	 /* LIST_REMOVE(ifp, if_list); */
	 goto error;
      }

      if (ifp == NULL)
	 goto error;
      
      /* 
       * Get IPv4 address 
       */
      if (ioctl(v, SIOCGIFADDR, ifr) < 0) {
	 goto error;
      } else {
	 Malloc(ifp->addr4, sockaddr_in);
	 if (ifp->addr4 == 0) {
	    fprintf(stderr, "   Out of memory\n");
	    exit(EX_CANTCREAT);
	 }
	 bzero((caddr_t)ifp->addr4, sizeof(struct sockaddr_in));
	 ifp->addr4->sin_addr.s_addr = 
	    ((struct sockaddr_in *)&ifr->ifr_addr)->sin_addr.s_addr;
	 ifp->addr4->sin_family = AF_INET;
      }
		
      /* 
       * IPv4 netmask 
       */
      if (ioctl(v, SIOCGIFNETMASK, ifr) < 0) {
	 goto error;
      }	else {
	 Malloc(ifp->mask4, sockaddr_in);
	 if (ifp->mask4 == 0) {
	    fprintf(stderr, "   Out of memory\n");
	    exit(EX_CANTCREAT);
	 }
	 bzero((caddr_t)ifp->mask4, sizeof(struct sockaddr_in));
	 ifp->mask4->sin_addr.s_addr = 
	    ((struct sockaddr_in *)&ifr->ifr_netmask)->sin_addr.s_addr;
	 ifp->mask4->sin_family = AF_INET;
      }
		
      /* 
       * Any IPv6 addr ? 
       */
      memcpy(ifr2.ifr_name, ifr->ifr_name, IFNAMSIZ);
      ifr2.ifr_addr.sa_family = AF_INET6;
      ifr2.ifr_len6 = sizeof(buf2);
      ifr2.ifr_buf6 = (caddr_t) &buf2;
      if (ioctl(u, SIOCGIFADDR, &ifr2) < 0) {
	 fprintf(stderr, "   Ioctl error\n");
	 goto error;
      } else {
	 if (ifr2.ifr_len6) {
	    fprintf(stderr,"   This host already has IPv6 addresses\n");
	    exit(EX_DATAERR);
	 }
      }
      LIST_INSERT_HEAD(&interfaces, ifp, if_list);
      continue;
   error:
   free(ifp); ifp = NULL;
   }
   
   if ((ifk_ieee == 0) && verbose)
      printf("\nNo IEEE LAN interface?\n");
   
   /* other interfaces */

   
}

/*
 * Determine the main LAN IEEE interface.
 * 3 ways to determine main interface:
 *	   - user has given us a name.
 *	   - only one ieee interface.
 *	   - from IPv4 default route.	
 */
struct interface *
find_main(char *name) {
   register struct interface *ifp;
   register int i;
	
   FILE *fp;              /* /proc/net/route file */
   unsigned int addr;     /* routes gateway addr  */
   unsigned char dev[10], /* device name */
                 line[256];   
   char *eof = NULL;
   
   /* 
    * user has provided a name 
    */
   if (name) {
      /* 
       * find matching struct 
       */
      for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
	 if (strcasecmp(ifp->name, name) == 0)
	    break;
      if (ifp == 0)
	 fprintf(stderr, "can't find interface %s\n", name);
      else if (ifp->kind != IFK_IEEE)
	 fprintf(stderr, "interface %s hasn't the good type (%d)?\n",
		 ifp->name, ifp->kind);
      else
	 return(ifp);
   }

   /* 
    * Easy case: 1 ieee interface 
    */
   if (ifk_ieee == 1) 
      for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
	 if (ifp->kind == IFK_IEEE)
	    return(ifp);
	
   /*
    * Get IPv4 default route 
    */
   fp = fopen("/proc/net/route", "r");
   do {
      eof = fgets(line, 256, fp);      
      i = sscanf(line, "%s %X %*X %*d %*d %*d %*d %*X %*d %*d %*d\n",
		 dev, &addr);
   } while ((eof) && (addr!= 0));
   fclose(fp);
   
   if (addr!= 0) { /* failure! */
      fprintf(stderr, "can't find default IPv4 route.\n");
      goto failed;
   }
   
   name = dev;
   
   for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
      if (strcasecmp(ifp->name, name) == 0)
	 break;
   
   if (ifp == 0) {
      fprintf(stderr, "IFP gives unknown interface %s\n", name);
      goto failed;
   }
   
   if (ifp->kind != IFK_IEEE) {
      if (verbose)
	 fprintf(stderr, "IFP gives bad type interface %s\n", ifp->name);
      goto failed;
   }
   return(ifp);
   
failed:
   for (ifp = interfaces.lh_first; ifp; ifp = ifp->if_list.le_next)
      if ((ifp->kind == IFK_IEEE) && (ifp->unit == 0))
	 return(ifp);
   
   /* failure! */
   fprintf(stderr, "can't find default IEEE interface?\n");
   exit(EX_OSFILE);
}

/*
 * Configure an IEEE LAN interface.
 */
void
setup_ieee(register struct interface *ifp) {
   register int cc;
   struct ifreq6 req6;
   struct ifreq  ifr;
   struct itimerval itv, itv2;
   
   if (verbose)
      printf("Setting up IEEE interface %s\n", ifp->name);
   
   /* 
    * IPv6 link-local addr 
    */
   if (Malloc(ifp->addr6, sockaddr_in6) == NULL) {
      fprintf(stderr, "out of memory\n");
      exit(EX_CANTCREAT);
   }
   bzero((caddr_t)ifp->addr6, sizeof(struct sockaddr_in6));
   ifp->addr6->sin6_family = AF_INET6;
   ifp->addr6->sin6_addr.s6_addr8[0] = 0xfe;
   ifp->addr6->sin6_addr.s6_addr8[1] = 0x80;
   /* EUI-64 Identifier */
   ifp->addr6->sin6_addr.s6_addr8[8] = ifp->laddr6[0];
   ifp->addr6->sin6_addr.s6_addr8[9] = ifp->laddr6[1];
   ifp->addr6->sin6_addr.s6_addr8[10] = ifp->laddr6[2];
   ifp->addr6->sin6_addr.s6_addr8[11] = 0xff;
   ifp->addr6->sin6_addr.s6_addr8[12] = 0xfe;
   ifp->addr6->sin6_addr.s6_addr8[13] = ifp->laddr6[3];
   ifp->addr6->sin6_addr.s6_addr8[14] = ifp->laddr6[4];
   ifp->addr6->sin6_addr.s6_addr8[15] = ifp->laddr6[5];   
 
   /* 
    * IPv6 mask (all 1s)
    */
   if (Malloc(ifp->mask6, sockaddr_in6) == NULL) {
      fprintf(stderr, "out of memory\n");
      exit(EX_CANTCREAT);
   }
   bzero((caddr_t)ifp->mask6, sizeof(struct sockaddr_in6));
   ifp->mask6->sin6_family = AF_INET6;
   memset((caddr_t)&ifp->mask6->sin6_addr, -1, 16);
 
   /* 
    * Setup interface with link-local address 
    */
   bzero((caddr_t)&req6, sizeof(req6));
   COPY_ADDR6(req6.in6, ifp->addr6->sin6_addr);
   req6.preflen = 16;
   ifr.ifr_buf6 = &req6;
   ifr.ifr_len6 = sizeof(req6);
   ifr.ifr_addr.sa_family = AF_INET6;
   bcopy((caddr_t)ifp->name, ifr.ifr_name, 6);
   
   if ((ioctl(u, SIOCSIFADDR, &ifr) < 0) &&
       !((ifk_ieee > 1) && (errno == EEXIST))) {
      perror("ioctl (SIOCSIFADDR)");
      fprintf(stderr, "can't setup interface %s\n", ifp->name);
      return;
   }
   
   if (verbose) {
      printf("Configuring %s %s/%d\n",
	     ifp->name,
	     inet_ntop(AF_INET6, &ifp->addr6->sin6_addr, inet_ntop_ret,
		       INET6_ADDRSTRLEN),
	     10 * 8);
      /*printf("%X:%X:%X:%X\n", ifp->addr6->sin6_addr.s6_addr32[0],
	ifp->addr6->sin6_addr.s6_addr32[1],
	ifp->addr6->sin6_addr.s6_addr32[2],
	ifp->addr6->sin6_addr.s6_addr32[3]);*/
   }
   
   /*
    * Perform duplicate address detection.
    */
   printf("Performing duplicate address detection.\n");
   ntosend = MAX_DUPL_ADDR_RETRANS + 1;
   ninflow = 0;
   
   /* Time before 1st probe is sent */
   delay = MAX_INITIALIZATION_DELAY * 1000000;
   delay -= (double)(random() % 1000000);
   itv.it_value.tv_sec = 0;
   itv.it_value.tv_usec = (u_int)delay;
   
   /* Time between probes */
   reload = DUPL_ADDR_RETRANS_TIMER * 1000000;  
   itv.it_interval.tv_sec = 0;
   itv.it_interval.tv_usec = reload;

   bzero((caddr_t)&itv2, sizeof(itv2));
   (void)setitimer(ITIMER_REAL, &itv, &itv2);
   
   for (;;) {
      cc = read(s, (char *)packet, MAXPACKET);
      if (cc < 0) {
	 if (errno != EINTR) {
	    perror("read (ICMP)");
	    continue;
	 }
      } else {
	 recv_pack((char *)packet, cc);
	 continue;
      }
      if (dring == 0) {
	 if (verbose)
	    printf("Read interrupt?\n");
	 continue;
      }
      if (--ntosend > 0) {
	 send_ndsol(ifp);
      }
      if (ntosend < 0) {
	 bzero((caddr_t)&itv, sizeof(itv));
	 bzero((caddr_t)&itv2, sizeof(itv2));
	 (void)setitimer(ITIMER_REAL, &itv, &itv2);
	 if (verbose)
	    printf("%s setup complete\n", ifp->name);
	 return;
      }
   }
}

/*
 * Add a default route for an interface.
 */ 
void
add_default_route(register struct interface *ifp, u_int flags) {
   struct in6_rtmsg rt;                   /* Netlink route message */
   struct in6_addr deflt = IN6ADDR_ANY_INIT;
   
   /*
    * Prepare message
    */
   bzero((char *)&rt, sizeof(rt));
   rt.rtmsg_type = SIOCADDRT;
   rt.rtmsg_flags = RTF_UP|flags;
   strcpy(rt.rtmsg_device, ifp->name);  
   /* Route is default ::/0 */
   COPY_ADDR6(rt.rtmsg_dst, deflt);
   rt.rtmsg_prefixlen = 0;     
        	
   if(ioctl(u, SIOCADDRT, &rt) < 0) {
      fprintf(stderr, "SIOCADDRT\n");
   }
   
   if (verbose) {
      printf("Added default route:\n\tFlags = %X\n", rt.rtmsg_flags);
      printf("\tdest = %s\n",
	     inet_ntop(AF_INET6, &deflt, inet_ntop_ret, INET6_ADDRSTRLEN));
		/*
	        printf("\tgateway = %s\n",
		       inet_ntop(AF_INET6, d, inet_ntop_ret,
				  INET6_ADDRSTRLEN));
	        printf("\tmask = %s\n",
		       inet_ntop(AF_INET6, d, inet_ntop_ret,
				  INET6_ADDRSTRLEN));*/
   } 
}


/*
 * Collision panic.
 */
void
collision(register struct interface *ifp) {
	/* Drastic ?! */
#ifdef notdef
   for (;;) {  
      (void) reboot(RB_HALT);
      perror("reboot!!!");
   }
#else
   for (;;)
      continue;
   /*
    * Notify the system administrator
    * Ask for a manual identifier.
    */
#endif
}







