#include "ndpd.h"


struct {
	struct cmsghdr c_cm;
	u_char c_space[64];
} c_cmsg;
#define cm	c_cmsg.c_cm

#define	MAXPACKET	8192	/* max packet size */
char packet[MAXPACKET];

int isv2;


/*--START FUNCTION--(send_rtsol)------------------------------------------

   Send an ICMPv6 Router Solicitation packet on ifp. 
   
   Last update:
   10/08/96: bb: Fixed checksum.
   08/12/96: bb

------------------------------------------------------------------------*/
void
send_rtsol(struct interface *ifp) {
   register struct ipv6hdr *ip;              /* header IPv6 */
   register struct r_sol *icp;               /* router sol. header */
   register struct nd_opt *hp;
   struct sockaddr_in6 dst;                  /* socket address */
   int icmp_len,                             /* ICMPv6 packet length */
      total_len,                             /* IPv6 packet length */
      cc;                                    /* sendto return code */
   unsigned int  csum;                       /* ICMPv6 checksum */
   struct timeout *tp;                       /* timeout structure */

#define LLA(x)	((struct nd_opt*)(x))   
   
   /*
    * Decrement sollicitation to send
    */
    
   if (ifp->ntosend >= 1)
      ifp->ntosend--;
   else
      return;

   /*
    * Destination address
    */
   bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
   dst.sin6_family = AF_INET6;
   /* All routers address on link local */
   dst.sin6_addr.s6_addr8[0] = 0xff;
   dst.sin6_addr.s6_addr8[1] = MADDR6_SCP_LINK;
   dst.sin6_addr.s6_addr8[15] = MADDR6_ALLROUTERS;

   if (verbose) {
      logtrace("Sending router sol for %s on %s\n",
	       inet_ntop(AF_INET6, &ifp->llocal, 
			 inet_ntop_ret, INET6_ADDRSTRLEN),
	       ifp->name);
   }

   /*
    * Packet size
    */
   icmp_len = sizeof(*icp) + 2 + (LTOKEN - 2); /* FIXME */
   total_len = icmp_len + sizeof(struct ipv6hdr);
   bzero(packet, total_len);

   /*
    * IPv6 header
    */
   ip = (struct ipv6hdr *)packet;
   ip->next_hdr = IPPROTO_ICMPV6;
   ip->payload = htons(icmp_len);

   COPY_ADDR6( ip->saddr, ifp->llocal);
   COPY_ADDR6(ip->daddr, dst.sin6_addr);

   /*
    * ICMPv6 header
    */
   icp = (struct r_sol *)(ip + 1);
   icp->hdr.type = ICMP6_SOLICITATION_RT;
   /* other fields are 0 */

   /*
    * Source link-layer address option
    */
   hp = (struct nd_opt *)(icp + 1);
   hp->type = ND_OPT_SLLA;
   hp->len  = 1;
   bcopy(ifp->token, LLA(hp)->lla, 3);    /* FIXME !!! */
   bcopy(ifp->token + 5, LLA(hp)->lla + 3, 3);

   /* 
    * Compute ICMPv6 checksum here 
    */
   csum = csum_partial((char *)icp, icmp_len, 0);
   icp->hdr.csum = csum_ipv6_magic(&ip->saddr, &ip->daddr, icmp_len, 
				   IPPROTO_ICMPV6, csum);

   /* 
    * Finish IPv6 header 
    */
   ip->vrprflow  = ICMP6_ND_PRIORITY | IPV6_VERSION;
   ip->hop_limit = ICMP6_ND_HOPS;

   if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		  &ifp->index,
		  sizeof(ifp->index)) < 0) {
      logperror("setsockopt (IPV6_MULTICAST_IF)");
      return;
   } 

   /*
    * Send it away!
    */
   cc = sendto(s, packet, total_len, 0,
	       (struct sockaddr *)&dst, sizeof (struct sockaddr_in6));

   if (cc < 0 || cc != total_len)  {
      logerr("wrote %d chars, ret=%d\n", total_len, cc);
      if (cc < 0) {
	 logperror("sendto");
	 return;
      }
   }
        
   /* 
    * No more sol. to send 
    */
   if (ifp->ntosend == 0) {
      /* if ifp->stateful, start stateful enquiry */
      return;
   }
       
   /* 
    * More sol. to send: set timeout 
    */
   if (Calloc(tp, timeout) == NULL) {
      logerr("out of memory\n");
      abort();
   }
   (void) gettimeofday(&now, NULL);
   tp->tv.tv_sec = now.tv_sec + RTR_SOLICITATION_INTERVAL;
   tp->tv.tv_usec = now.tv_usec;
   tp->func = send_rtsol;
   tp->arg = ifp;
   add_timeout(tp);
}

/*--START FUNCTION--(recv_icmp)--------------------------------------------

   Receive an ICMPv6 (Router Advertisement or Redirect) packet. 
   
   Last update:
   10/08/96: bb

---------------------------------------------------------------------------*/
void
recv_icmp() {
   register struct ipv6hdr *ip;              /* IPv6 header */
   register struct r_red *redirect;
   struct interface *ifp;                    /* Interface structure */
   struct msghdr msg;                        /* Message header (from kernel) */
   struct iovec iov;                         /* Iovec in message */
   int cc;                                   /* recvmsg return value */

   /*
    * Initializes message structure
    */
   bzero((caddr_t)&msg, sizeof(struct msghdr));
   bzero((caddr_t)&iov, sizeof(struct iovec));
   msg.msg_iov = &iov;
   msg.msg_iov->iov_base = packet;
   msg.msg_iov->iov_len = MAXPACKET;
   msg.msg_iovlen = 1;
   msg.msg_control = (caddr_t)&c_cmsg;
   msg.msg_controllen = sizeof(c_cmsg);

   /*
    * Receive message
    */
   if ((cc = recvmsg(s, &msg, 0)) < 0) {
      if (verbose)
	 logperror("read (ICMPv6)");
      return;
   }
   
   /* error on reception */
   if (cc == 0) {
      if (verbose)
	 logerr("read ICMPv6 return 0\n");
      return;
   }
   
   /* message truncated */
   if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) {
      if (verbose)
	 logerr("read ICMPv6 truncated\n");
      return;
   }
   
   /* control error */
   if ((msg.msg_controllen == 0) ||
       (cm.cmsg_len != (sizeof(struct cmsghdr) + 
			sizeof(struct in6_pktinfo))) ||
       (cm.cmsg_level != IPPROTO_IPV6) ||
       (cm.cmsg_type != IPV6_PKTINFO)) {
      if (verbose)
	 logerr("read ICMPv6 doesn't give interface %d %d %d %d %d\n",
		msg.msg_controllen,
		IPPROTO_IPV6, cm.cmsg_level,
		IPV6_PKTINFO, cm.cmsg_type
		);
      return;
   }

   /*
    * Look for concerned interface
    */
   for (ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next)
      if (ifp->index ==
	  ((struct in6_pktinfo *)cm.cmsg_data)->ipi6_ifindex) 
	 break;
   if (ifp == NULL) {
      if(verbose)
	 logerr("No corresponding interface handled, %d\n",
		((struct in6_pktinfo *)cm.cmsg_data)->ipi6_ifindex);
      return;
   }	
#ifdef NDPD_HOST_DEBUG
   dump_hex((char *)packet, 40);
#endif
   
   ip = (struct ipv6hdr *)packet;
   if (verbose) {
      logtrace("******* in IPV6 *******\n"
	       "vrprflow :\t%4X\n"
	       "payload  :\t%d\n"
	       "hop_limit:\t%d\n"
	       "nxt_hdr  :\t%d\n"
	       "saddr    :\t%s\n"
	       "daddr    :\t%s\n",
	       ip->vrprflow, ntohs(ip->payload), ip->hop_limit,
	       ip->next_hdr,
	       inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret, 
			 INET6_ADDRSTRLEN), 
	       inet_ntop(AF_INET6, &ip->daddr, inet_ntop_ret1, 
			 INET6_ADDRSTRLEN));
   }
   
   /* message too short */
   if (cc < sizeof(*ip) + ICMP6_MINLEN) {
      if (verbose)
	 logerr("ICMPv6: too short (%d)\n", cc);
      return;
   }
#ifdef NDPD_HOST_DEBUG	
   dump_hex((char *)(ip+1), sizeof(struct r_adv)+ sizeof(struct nd_opt)+ sizeof(struct pfx_opt));
#endif
   
   /* non saddr address */
   if (!IS_LINKLADDR6(ip->saddr)) {
      if (verbose)
	 logerr("Source address isn't a link address:\n%s\n",
		cc, 
		inet_ntop(AF_INET6, &ip->saddr,
			  inet_ntop_ret,INET6_ADDRSTRLEN));
      return;
   }

   /* wrong hop limit */
   if (ip->hop_limit != ICMP6_ND_HOPS)
      return;
   /* FIXME: priority must be 15 */

   cc -= sizeof(struct ipv6hdr);
   redirect = (struct r_red *)(ip + 1);
        
   /*
    * Received redirect msg.
    */
   if (redirect->hdr.type == ICMP6_REDIRECT) {
      if (SAME_ADDR6(redirect->tgt, redirect->addr))/* Is on local link */
	 recv_llredirect(ifp, ip, cc);
      else                                     /* Usual redirect */
	 recv_rtredirect(ifp, ip, cc);
   } else
      /*
       * Received router advertisement
       */
      if (redirect->hdr.type == ICMP6_ADVERTISMENT_RT)
	 recv_rtadv(ifp, ip, cc);
   return;
}
/*--END FUNCTION--(recv_icmp)-----------------------------------------------*/

/*--START FUNCTION--(recv_rtadv)--------------------------------------------

  Receives a router advertisement
  Last update:
  08/12/96: bb

---------------------------------------------------------------------------*/
void
recv_rtadv(struct interface *ifp, 
	   register struct ipv6hdr *ip, 
	   int cc) {
   register struct r_adv *hp;               /* Router advertisement header */
   register struct pfx_opt *po;             /* Prefix option */
   register struct nd_opt *no;              /* ND option */

   struct prefix *pp;
   struct router *rp;
   unsigned char *uc;
   int changed = 0;

#define LLA(x)	((struct nd_opt*)(x))
#define PREF(x)	((struct pfx_opt *)(x))

   /*
    * Router advertisement header
    */
   hp = (struct r_adv *)(ip + 1);
   
   /* Invalid code */
   if (hp->hdr.code != 0) {
      if (verbose)
	 logerr("ICMPv6 Rt Adv: code=%d\n", hp->hdr.code);
      return;
   }
   
   /* Bad length (too short) */
   if ((unsigned)cc < ICMP6_RALEN) {
      if (verbose)
	 logerr("ICMPv6 Rt Adv: too short %d\n", cc);
      return;
   }

   if (verbose) {
      logtrace("receive ICMPv6 Rt Adv on %s from %s",
	       ifp->name,
	       inet_ntop(AF_INET6, &ip->saddr,
			 inet_ntop_ret,INET6_ADDRSTRLEN));
      if (isv2)
	 logtrace(" (draftV2)");
      logtrace(" to %s (len=%d)\n",
	       inet_ntop(AF_INET6, &ip->daddr,
			 inet_ntop_ret,INET6_ADDRSTRLEN),
	       cc + sizeof(struct ipv6hdr));
   }
   
   /* 
    * Needn't advertise any more 
    */
   ifp->ntosend = 0;

   /* RFC 1970 - 6.3.4 
    * Timer set in update_router
    * Find router in interface router list
    */
   for (rp = BEST(ifp); rp; rp = rp->rt_list.tqe_next)
      if (SAME_ADDR6(rp->addr.sin6_addr, ip->saddr))
	 break;
   
   /*
    * No router entry 
    */
   if (rp == NULL) {
      if (verbose)
	 logtrace("allocate router\n");

      /*
       * Alloc new router entry
       */
      if (Calloc(rp, router) == NULL) {
	 if (verbose)
	    logerr("out of memory\n");
	 return;
      }
      
      rp->ifp = ifp;
      rp->addr.sin6_family = AF_INET6;
      COPY_ADDR6(rp->addr.sin6_addr, ip->saddr);
      strcpy(rp->phy, ifp->name);
      /* ???? */
      strncpy(rp->token, ifp->token, LTOKEN); 
      TAILQ_INSERT_TAIL(&ifp->routers, rp, rt_list);
   } 	
   
   /* 
    * Lifetime
    */
   rp->lifetime = ntohs(hp->lifetime);
	
   /*
    * Flags
    */
   uc = (unsigned char *)&hp->flags;
   rp->nflags = *uc;
   rp->nflags &= (RTADV_FLG_M|RTADV_FLG_O);
	
   /* 
    * Hop limit always 255 
    */
	
   /* 
    * Reachable time 
    */
	
   /* 
    * Retrans timer 
    */
      
   
   if (verbose)
      logtrace("Router %s %x %d\n",
	       inet_ntop(AF_INET6, &rp->addr.sin6_addr,
			 inet_ntop_ret, INET6_ADDRSTRLEN),
	       *uc, (int)rp->lifetime);

   /*
    * Read options
    */
   cc -= sizeof(*hp);
   no = (struct nd_opt *)(hp + 1);
   while (cc) 
      switch (no->type) {
	 /* 
	  * Source Link Layer Address Option
	  * Kernel should take care of creating ND entry.
	  */
      case ND_OPT_SLLA:
	 if (verbose)
	    printf("ND_OPT_SLLA:\n");
	 
	 if (no->len != (((LTOKEN-2) + 7)/8)) {
	    if (verbose)
	       logerr("Bad link-layer address length (%d)\n",
		      LLA(hp)->len << 3);
	    return;
	 }
	 
	 if (cc < (no->len << 3)) {
	    if (verbose)
	       logerr("Rt Adv underflow\n");
	    return;
	 }
	 
	 rp->alen = LTOKEN - 2; /* FIXME... ethernet only */
	 bcopy(no->lla, rp->phy, LTOKEN-2);
	 goto next;

      case ND_OPT_MTU:
	 /* 
	  * See kernel code 
	  */
	 break;
	 
	 /*
	  * Prefix information
	  */
      case ND_OPT_PREFIX_INFO:
	 if (verbose)
	    printf("ND_OPT_PREFIX_INFO:\n");
	 
	 po = (struct pfx_opt *) no;
	 
	 /* Bad length */
	 if (po->len != 4) {
	    if (verbose)
	       logerr("bad PREF length (%d)\n",
		      po->len << 3);
	    return;
	 }
	 
	 if (cc < (po->len << 3)) {
	    if (verbose)
	       logerr("Rt Adv underflow\n");
	    goto next;
	 }
	 
	 if (po->plen > 128) {
	    if (verbose)
	       logerr("Bad PREF plen (%d)\n",
		      PREF(hp)->plen);
	    return;
	 }
	 
	 if (po->plen != (128-8*LTOKEN)) {
	    if (verbose)
	       logerr("Strange PREF plen (%d)\n", po->plen);
	    goto next;
	 }
	 
	 /* Bad prefix */
	 if ((po->prefix.s6_addr8[0] == 0) ||
	     (po->prefix.s6_addr8[0] == 0xff) ||
	     IS_LINKLADDR6(po->prefix)) {
	    if (verbose)
	       logerr("bad PREF prefix %s\n",
		      inet_ntop(AF_INET6, &po->prefix,
				inet_ntop_ret, INET6_ADDRSTRLEN));
	    goto next;
	 }
	 
	 /*
	  * No autoconf for this interface
	  */
	 if (!ifp->autoconf)
	    goto next;
	 
	 /* 
	  * Lifetimes 
	  */
	 po->valid_lt = ntohl(po->valid_lt);
	 po->pref_lt = ntohl(po->pref_lt);

	 if (verbose)
	    logtrace("Prefix %s/%d %x (%u/%u)\n",
		     inet_ntop(AF_INET6, &po->prefix,
			       inet_ntop_ret, INET6_ADDRSTRLEN),
		     po->plen,
		     po->flags,
		     po->valid_lt,
		     po->pref_lt);
		
	 /* if (isv2 && po->valid_lt == 0)
	    po->valid_lt = po->pref_lt = INFLIFE; */
		
	 for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next) {
	    logtrace("%s\n",
		     inet_ntop(AF_INET6, &po->prefix,
			       inet_ntop_ret, INET6_ADDRSTRLEN));
	    logtrace("%s\n",
		     inet_ntop(AF_INET6, &pp->pref,
			       inet_ntop_ret, INET6_ADDRSTRLEN));
	    logtrace("%d/%d\n",  po->plen, pp->plen);   
	    if (SAME_ADDR6(po->prefix, pp->pref) &&
		(po->plen == pp->plen)) 
	       /* prefix already in list */
	       break;
	 }
	 
	 /*
	  * No such prefix entry
	  */
	 if (pp == NULL) {
#define ND_PREF_FLAGS	(ND_PREF_FLG_L|ND_PREF_FLG_A)
	    if (!po->flags.onlink && !po->flags.auto1)
	       goto next;

	    if (verbose)
	       logtrace("allocate prefix\n");

	    /*
	     * Alloc new prefix entry
	     */
	    if (Calloc(pp, prefix) == NULL) {
	       if (verbose)
		  logerr("out of memory\n");
	       return;
	    }
	    pp->ifp = ifp;
	    COPY_ADDR6(pp->pref, po->prefix);
	    pp->plen = po->plen;
	    pp->pflags = 0;
	    mk_pref(pp);
	    mk_addr(pp);
	    if (FIRST(ifp) == NULL)
	       COPY_ADDR6(ifp->addr, pp->addr);
	    TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
	 }
		
	 
	 uc = (unsigned char *)&po->flags;
	 pp->nflags = *uc;
	 printf("FLAGS: %X\n", pp->nflags);
	 ((u_char)pp->nflags) &= ND_PREF_FLAGS;
	 printf("FLAGS: %X\n", pp->nflags);
	 pp->nflags |= 1;
	 pp->deplife = po->pref_lt;
	 pp->invlife = po->valid_lt;
	 if (pp->deplife == 0)
	    goto next;
	 if (pp->invlife == INFLIFE)
	    goto next;
	 if (pp->invlife < pp->deplife)
	    pp->invlife = pp->deplife;
	 goto next;

      default:
      next:	
#ifdef NDPD_HOST_DEBUG		
	 printf("default-next: no->len = %d; type = %d\n",
		no->len, no->type);
#endif
	 if (no->len == 0) {
	    if (verbose)
	       logerr("null option length (%d)\n", no->type);
	    return;
	 }
	 cc -= no->len << 3;
#ifdef NDPD_HOST_DEBUG
	 printf("cc = %d\n", cc);
#endif
	 if (cc < 0) {
	    if (verbose)
	       logerr("Rt Adv underflow\n");
	    return;
	 }
	 if (cc == 0)
	    break;
	 no = (struct nd_opt *)((char *)no + (no->len << 3));
      }
	
   /*
    * Update prefix status
    */
   for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next)
      if (pp->pflags != pp->nflags) {
	 update_prefix(pp);
	 changed++;
      }
   if (changed)
      sort_prefix(ifp);
   update_router(rp);
}

/*--START FUNCTION--(recv_llredirect)------------------------------------
  
  Called when an incoming redirect message indicates that a given host
  is on-link
  
  ----------------------------------------------------------------------*/
void
recv_llredirect(struct interface *ifp, struct ipv6hdr *ip, int cc) {
   register struct r_red *icp;
   struct router *rp;
   struct nd_opt *xp;
   char *mac = NULL;
   
   /* Proc search */
   int wcount;
   FILE *fd;
   char line[255], dev_name[IFNAMSIZ], *fgets_ret_val;
   u_long mtu;
   unsigned short flags;
   struct in6_addr dst6;
   /* Route update */
   struct in6_rtmsg rt6;
   
   
   icp = (struct r_red *)(ip + 1);

   if (icp->hdr.code != 0) {
      if (verbose)
	 logerr("ICMPv6 llRedirect: code=%d\n", icp->hdr.code);
      return;
   }
   if ((unsigned)cc < (sizeof(struct r_red))) {
      if (verbose)
	 logerr("ICMPv6 llRedirect: too short %d\n", cc);
      return;
   }

   for (rp = BEST(ifp); rp; rp = rp->rt_list.tqe_next)
      if (SAME_ADDR6(icp->tgt, rp->addr.sin6_addr)) {
	 if (verbose)
	    logtrace("link-layer redirect for router %s\n",
		     inet_ntop(AF_INET6, &icp->addr,
			       inet_ntop_ret, sizeof(struct in6_addr)));
	 get_rtphy(rp);
	 return;
      }
      
   /* Router was not found in list */ 
   cc -= sizeof(struct r_red);
   xp = (struct nd_opt *)((caddr_t)icp + sizeof(struct r_red));
      
   /* Retrieve target's MAC address */
   while (cc && (mac == NULL)) 
      switch (xp->type) {
      case ND_OPT_TLLA:
	 if (xp->len < 1) {
	    if (verbose)
	       logerr("bad LL address length (%d)\n", xp->len << 3);
	    return;
	 }
	 if (cc < (xp->len << 3)) {
	    if (verbose)
	       logerr("llRedirect underflow\n");
	    return;
	 }
	 mac = (char *)xp->lla;
	 break;

      default:
	 if (xp->len == 0) {
	    if (verbose)
	       logerr("null option length (%d)\n", xp->type);
	    return;
	 }
	 cc -= xp->len << 3;
	 if (cc < 0) {
	    if (verbose)
	       logerr("llRedirect underflow\n");
	    return;
	 }
	 xp = (struct nd_opt *)((caddr_t)xp + (xp->len << 3));
      }
   if (mac == NULL) {
      if (verbose)
	 logerr("llRedirect without LL address\n");
      return;
   }

   if (verbose) {
      logtrace("link-layer redirect for dest %s",
	       inet_ntop(AF_INET6, &icp->addr, inet_ntop_ret,
			 sizeof(struct in6_addr)));
   }
        
   /* 
    * Get old route to redirected destination
    */
   fd = fopen("/proc/net6/route6", "r");
   do {
      fgets_ret_val = fgets(line, 255, fd);
      wcount = sscanf(line,
		      "%s %X:%X:%X:%X %*X:%*X:%*X:%*X %4X %*d %*d %*d %*d %lu %*d %*u",
		      dev_name, 
		      &dst6.s6_addr32[0],
		      &dst6.s6_addr32[1],
		      &dst6.s6_addr32[2],
		      &dst6.s6_addr32[3],
		      & flags,
		      &mtu);
      if ((wcount = 7)||(SAME_ADDR6(icp->addr, dst6)))
	 break;
   } while (fgets_ret_val);
   fclose(fd);
   if (fgets_ret_val == NULL) {
      if (verbose)
	 logerr("Cannot retrieve original route!\n");
      return;
   }
	
   /*
    * Kernel returns 
    */
   if ((flags & RTF_GATEWAY) == 0) {
      if (verbose)
	 logtrace("link-layer redirect for a direct route\n");
      return;
   }

   if (mtu < IPV6_MTU)
      mtu = ifp->mtu;
	
   /* 
    * Update route to redirected destination 
    */
   /* Does not use netlink for now */
   memset(&rt6, 0, sizeof(rt6));
   rt6.rtmsg_type      = SIOCADDRT;
   COPY_ADDR6(rt6.rtmsg_dst, dst6);
   rt6.rtmsg_prefixlen = 128; /* local */
   strcpy(rt6.rtmsg_device, dev_name);
   rt6.rtmsg_flags     = RTF_UP | RTF_HOST | RTF_MODIFIED;	
	             
   if (ioctl(u, SIOCADDRT, &rt6) < 0) {
      if (verbose)
	 logerr("Socket error: can't redirect route.");
      return;
   }
}

/*--START FUNCTION--(recv_rtredirect)----------------------------------------
  
  ------------------------------------------------------------------------*/
void
recv_rtredirect(struct interface *ifp, struct ipv6hdr *ip, int cc) {
   register struct r_red *icp;

   /* Proc search */
   FILE *fd;
   int wcount;
   char line[255], dev_name[IFNAMSIZ], *fgets_ret_val;
   u_long  mtu;
   u_short flags;
   struct in6_addr dst6, gw6;
   /* Route update */
   struct in6_rtmsg rt6;
	

   icp = (struct r_red *)(ip + 1);

   if (icp->hdr.code != 0) {
      if (verbose)
	 logerr("ICMPv6 rtRedirect: code=%d\n", icp->hdr.code);
      return;
   }
   if ((unsigned)cc < sizeof(struct r_red)) {
      if (verbose)
	 logerr("ICMPv6 rtRedirect: too short %d\n", cc);
      return;
   }

   if (verbose) {
      logtrace("router redirect for dest %s",
	       inet_ntop(AF_INET6, &icp->addr, inet_ntop_ret,
			 sizeof(struct in6_addr)));
      logtrace(" via new gateway %s\n",
	       inet_ntop(AF_INET6, &icp->tgt, inet_ntop_ret,
			 sizeof(struct in6_addr)));
   }
        
   /* 
    * Get old route to redirected destination (through /proc)
    */
   fd = fopen("/proc/net6/route6", "r");
   do {
      fgets_ret_val = fgets(line, 255, fd);
      wcount = sscanf(line,
		      "%s %X:%X:%X:%X %X:%X:%X:%X %4X %*d %*d %*d %*d %lu %*d %*u",
		      dev_name, 
		      &dst6.s6_addr32[0],
		      &dst6.s6_addr32[1],
		      &dst6.s6_addr32[2],
		      &dst6.s6_addr32[3],
		      &gw6.s6_addr32[0],
		      &gw6.s6_addr32[1],
		      &gw6.s6_addr32[2],	   
		      &gw6.s6_addr32[3],	   
		      &flags,
		      &mtu);
      if ((wcount = 11)&&
	  (SAME_ADDR6(icp->addr, dst6))&&
	  (SAME_ADDR6(ip->saddr, gw6)))
	 break;
   } while (fgets_ret_val);
   fclose(fd);
   if (fgets_ret_val == NULL) {
      if (verbose)
	 logerr("Cannot retrieve original route!\n");
      return;
   }
	
   (void)gettimeofday(&now, NULL);

   /* Check retrieved route */
   if ((flags & RTF_GATEWAY) == 0) {
      if (verbose)
	 logerr("router redirect for a direct route\n");
      return;
   }
   if (strcmp(dev_name, ifp->name)) {
      if (verbose)
	 logerr("bad interface name (%s != %s)\n",
		dev_name, ifp->name);
      return;
   }

   if (mtu < IPV6_MTU)
      mtu = ifp->mtu;

   /* 
    * Now, change/modify route
    */
   memset(&rt6, 0, sizeof(rt6));
   rt6.rtmsg_type      = SIOCADDRT;
   COPY_ADDR6(rt6.rtmsg_dst, dst6);
   COPY_ADDR6(rt6.rtmsg_gateway, gw6);
   rt6.rtmsg_prefixlen = 128;  
   strcpy(rt6.rtmsg_device, dev_name);
   rt6.rtmsg_flags     = flags | RTF_MODIFIED;	
	             
   if (ioctl(u, SIOCADDRT, &rt6) < 0) {
      if (verbose)
	 logerr("Socket error: can't redirect route.");
      return;
   }
}













