#include "ndpd.h"
extern void dump_hex(char *buf, int nb);
/*-----------------------------------------------------------------------
  recv_route()
  init()
 -----------------------------------------------------------------------*/

/*--START FUNCTION--(recv_route)------------------------------------------
  
  
 -----------------------------------------------------------------------*/
void
recv_route()
{
   register struct interface *ifp;
   struct prefix *pp;
   struct timeout *tp;
   
   struct netlink_rtinfo *nrt;
   struct sockaddr_in6 *sin, *sin_mask;
   
   long delay;
   int cc;                    /* Read call return value */
   char cp[2048];
   u_char plen = 0;

   /*
    * Read kernel route message
    */
   bzero(cp, 2048);
   cc = read(r, cp, 2048);
   if (cc <= 0) {
      if (verbose)
	 logperror("read (route message)");
      return;
   } 
   
   if (cc != sizeof(struct netlink_rtinfo)) {
      if (verbose)
	 logerr("read (route message) bad length.\n");
      return;
   }
   
   /*
    * Read through structure
    */
   nrt = (struct netlink_rtinfo *) cp;
   if (verbose)
      logtrace("receive route (%d,%s,%x)\n",
	       nrt->rtmsg_type, 
	       nrt->rtmsg_device, 
	       nrt->rtmsg_flags);

   switch (nrt->rtmsg_type) {
      /*
       * New address on host
       */
   case RTMSG_NEWADDR:
      for (ifp = interfaces.tqh_first;
	   ifp && (strcmp(nrt->rtmsg_device, ifp->name)!=0);
	   ifp = ifp->if_list.tqe_next)
	 continue;
      if ((ifp == NULL) || !ifp->autoconf)
	 return;
	        
      /* New address */
      sin = (struct sockaddr_in6 *)&nrt->rtmsg_dst6;
      
      /* Validity check */
      if ((sin == NULL) ||
	  IS_ANYADDR6(sin->sin6_addr) ||
	  IS_LINKLADDR6(sin->sin6_addr) ||
	  (sin->sin6_addr.s6_addr8[0] == 0) ||
	  (sin->sin6_addr.s6_addr8[0] == 0xff))
	 return;
      if (sin->sin6_family != AF_INET6)
	 return;
      
      /* Mask */
      sin_mask = (struct sockaddr_in6 *)&nrt->rtmsg_gw6;  
      if (sin->sin6_family != AF_INET6)
	 return;
      
      /* Prefix length */
      plen = nrt->rtmsg_plen6;
      if (plen < 3)
	 return;
		
      if (verbose)
	 logtrace("Get new address %s/%d on %s\n",
		  inet_ntop(AF_INET6, &sin->sin6_addr, 
			    inet_ntop_ret,INET6_ADDRSTRLEN),
		  plen, ifp->name);
		
      /* Prefix already in list? */
      for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next) {
	 logtrace("%s\n",
		  inet_ntop(AF_INET6, &sin->sin6_addr,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	 logtrace("%s\n",
		  inet_ntop(AF_INET6, &pp->addr,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	 logtrace("%d/%d\n",  plen, pp->plen);   
	 if (SAME_ADDR6(sin->sin6_addr, pp->addr) &&
	     (plen == pp->plen)) 
	    /* prefix already in list */
	    break;
      }
	
      /* 
       * No... new prefix entry... 
       */
      if (pp == NULL) {
	 if (Calloc(pp, prefix) == NULL) {
	    if (verbose)
	       logerr("out of memory\n");
	    return;
	 }

	 pp->plen = plen;
	 COPY_ADDR6(pp->addr, sin->sin6_addr);
	 COPY_ADDR6(pp->pref, pp->addr);
	 mk_pref(pp);
	 mk_mask(&pp->amask, (int)plen);
	 pp->ifp = ifp;
	 pp->invlife = INFLIFE;
	 pp->deplife = DEFDLIFE;
	 pp->nflags = ND_PREF_FLG_A;
	 if (pp->plen != 128)
	    pp->nflags |= ND_PREF_FLG_L;
	 pp->pflags = pp->nflags;
	 TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
	 update_prefix(pp);
	 sort_prefix(ifp);
      }
      return;

      /*
       * Address deleted on host
       */
   case RTMSG_DELADDR:
      /* Which interface ? */
      for (ifp = interfaces.tqh_first;
	   ifp && (strcmp(ifp->name, nrt->rtmsg_device) != 0);
	   ifp = ifp->if_list.tqe_next)
	 continue;
      if (ifp == NULL)
	 return;
	        
      sin = (struct sockaddr_in6 *)&nrt->rtmsg_dst6;
      if ((sin == NULL) ||
	  IS_ANYADDR6(sin->sin6_addr) ||
	  IS_LINKLADDR6(sin->sin6_addr) ||
	  (sin->sin6_addr.s6_addr8[0] == 0) ||
	  (sin->sin6_addr.s6_addr8[0] == 0xff))
	 return;
      
      if (sin->sin6_family != AF_INET6)
	 return;
      
      sin_mask = (struct sockaddr_in6 *)&nrt->rtmsg_gw6;  
      if (sin_mask->sin6_family != AF_INET6)
	 return;
      
      plen = nrt->rtmsg_plen6;
      if (plen < 3)
	 return;
		
      for (pp = ifp->prefixes.tqh_first;
	   pp && ((pp->plen != plen) || !SAME_ADDR6(pp->addr, sin->sin6_addr));
	   pp = pp->prf_list.tqe_next)
	 continue;
      
      if (pp == NULL)
	 return;
      
      prefix_inval(pp);
      return;

   case RTMSG_NEWDEVICE:
   case RTMSG_DELDEVICE:
      for (ifp = interfaces.tqh_first;
	   (ifp && 
	    ((strcmp(ifp->name, nrt->rtmsg_device)!=0) ||
	     (((ifp->flags ^ nrt->rtmsg_flags) & IFF_UP) == 0)));
	   ifp = ifp->if_list.tqe_next)
	 continue;
      if (ifp == NULL)
	 return;

      ifp->flags = nrt->rtmsg_flags;
      if ((ifp->flags & IFF_UP) == 0) {
	 if (verbose)
	    logtrace("interface %s down\n", ifp->name);
	 if (ifp == def_if)
	    reset_default();
	 return;
      }

      if (verbose)
	 logtrace("interface %s up\n", ifp->name);

      (void) gettimeofday(&now, NULL);

      ifp->ntosend = MAX_RTR_SOLICITATIONS;
      if (Calloc(tp, timeout) == NULL) {
	 logerr("out of memory\n");
	 abort();
      }
      delay = MAX_RTR_SOLICITATION_DELAY * 1000000;
      delay += (random() % 1000000) - 499999;
      tp->tv.tv_sec = now.tv_sec + delay / 1000000;
      tp->tv.tv_usec = now.tv_usec + delay % 1000000;
      NORMTIME(tp->tv);
      tp->func = send_rtsol;
      tp->arg = ifp;
      add_timeout(tp);

      if (Calloc(tp, timeout) == NULL) {
	 logerr("out of memory\n");
	 abort();
      }
      tp->tv.tv_sec = now.tv_sec + 5;
      tp->tv.tv_usec = now.tv_usec;
      tp->func = reset_default;
      add_timeout(tp);
      return;
            
      /* FIXME
	 case RTM_RTLOST:
	 #define RTA_RTLOST	(RTA_DST|RTA_GATEWAY)
	 if ((rtm.rtm_addrs & RTA_RTLOST) != RTA_RTLOST)
	 return;
	 cp += sizeof(struct rt_msghdr);
	 sin = (struct sockaddr_in6 *)cp;
	 if (sin->sin6_family != AF_INET6)
	 return;
	 ADVANCE(cp, sin);
	 sdl = (struct sockaddr_dl *)cp;
	 if ((sdl->sdl_family != AF_LINK) ||
	 (sdl->sdl_index == 0))
	 return;
	 bcopy(sin, &newsin, sizeof(*sin));
	 sin = &newsin;
	 bcopy(sdl, &newsdl, sizeof(*sdl));
	 sdl = &newsdl;
	 router_lost(sin);
	 for (ifp = interfaces.tqh_first;
	 ifp && (ifp->index != sdl->sdl_index);
	 ifp = ifp->if_list.tqe_next)
	 continue;
	 if (ifp == NULL)
	 return;
	 if (verbose)
	 logtrace("%s router %s on %s\n",
	 sdl->sdl_alen != 0 ? "lost" : "unreachable",
	 inet_ntop(AF_INET6, &sin->sin6_addr, NULL,
	 sizeof(struct in6_addr)),
	 ifp->name);
	 for (rp = ifp->routers.tqh_first; rp;
	 rp = rp->rt_list.tqe_next)
	 if (SAME_ADDR6(rp->addr.sin6_addr, sin->sin6_addr))
	 break;
	 if (rp == NULL)
	 return;
	 if (sdl->sdl_alen == 0)
	 router_unreachable(rp);
	 else
	 router_invalid(rp);
	 return;
	 */   
		
   default:
      return;		
   }
}

/*--START FUNCTION--(if_fetch)-----------------------------------------------
  
  
----------------------------------------------------------------------------*/
static int if_fetch(char *ifname, struct interface *ife)
{
   struct ifreq ifr;
 
   /* 
    * Name
    */
   strcpy(ife->name, ifname);

   /*
    * Flags 
    */
   strcpy(ifr.ifr_name, ifname);
   if (ioctl(u, SIOCGIFFLAGS, &ifr) < 0) {
      fprintf(stderr, "Can't get interface %s 's flags.", ifname);
      return(-1);
   }
#define IFF_LAN (IFF_UP|IFF_BROADCAST|IFF_MULTICAST)
   if (((ifr.ifr_flags & IFF_LAN) != IFF_LAN)) {
      fprintf(stderr, "%s: Not a LAN interface\n", ifname);  
      return -1;
   } 
   ife->flags = ifr.ifr_flags;
    
   /*
    * Index 
    */
   if(ioctl(u, SIOGIFINDEX, &ifr) < 0){
      fprintf(stderr, "Can't get interface %s's index.", ifname);
      return (-1);
   }
   ife->index = ifr.ifr_ifindex;

   /* 
    * Token  (EUI 64)
    */
   strcpy(ifr.ifr_name, ifname);
   if (ioctl(u, SIOCGIFHWADDR, &ifr) < 0) {
      fprintf(stderr, "Can't get token for %s\n", ife->name);
      return -1; /* No token */
   } 
   ife->token[0] = ifr.ifr_hwaddr.sa_data[0];
   ife->token[1] = ifr.ifr_hwaddr.sa_data[1];
   ife->token[2] = ifr.ifr_hwaddr.sa_data[2];
   ife->token[3] = 0xff;
   ife->token[4] = 0xfe;
   ife->token[5] = ifr.ifr_hwaddr.sa_data[3];
   ife->token[6] = ifr.ifr_hwaddr.sa_data[4];
   ife->token[7] = ifr.ifr_hwaddr.sa_data[5];
    
   /* 
    * MTU 
    */
   strcpy(ifr.ifr_name, ifname);
   if ((ioctl(u, SIOCGIFMTU, &ifr) < 0) || (ifr.ifr_mtu < IPV6_MTU)) {
      fprintf(stderr, "Can't get valid PMTU for %s\n", ife->name);
      return -1; /* Invalid MTU */
   }
   ife->mtu = ifr.ifr_mtu;
   return(0);
}

/*-------------------------------------------------------------------------
  Initializes daemon lists.
  
  ---------------------------------------------------------------------*/
void
init() {
   struct ifconf ifc;
   register struct ifreq *ifr1;
   struct ifreq ifr;

   register struct interface *ifp;
   struct interface *ifp_tmp;
   register struct prefix *pp;
   int i, 
      itf_addr_nb;
   
   char *buf, *buf2;
   char buff[1024], buff2[1024];

   /*
    * Init lists
    */
   TAILQ_INIT(&interfaces);
   TAILQ_INIT(&timeouts);
   ifp = NULL;

   /*
    * Get all interfaces
    */
   bzero((caddr_t)&ifc, sizeof(ifc));
   bzero((caddr_t)&ifr, sizeof(ifr));

   ifc.ifc_len = sizeof(buff);
   ifc.ifc_buf = buff;
   ifr.ifr_addr.sa_family = AF_INET6;
   memcpy(buff, ((char *) &ifr), sizeof(struct ifreq));
   if (ioctl(u, SIOCGIFCONF, &ifc) < 0) {
      logperror("ioctl0");
      exit(4);
   }
   
   /*
    * Scan interface list returned by kernel
    */
   buf = buff;
   i = ifc.ifc_len / sizeof(struct ifreq);

   for (; i > 0; i--,  buf+=sizeof(struct ifreq)){
      ifr1 = (struct ifreq *)buf;
      
      /*
       * New element in application interface list
       */
      if (Calloc(ifp, interface) == NULL) {
	 logerr(stderr, "out of memory\n");
	 abort();
      }
      memset((char *) ifp, 0, sizeof(struct interface));
      
      /*
       * Get interface parameters
       */
      if (if_fetch(ifr1->ifr_name, ifp) < 0) {
	 fprintf(stderr, "%s: Will not handle this interface\n",
		 ifr1->ifr_name);
	 free(ifp);
	 continue;
      }
                
      /* 
       * Queue tail 
       */
      TAILQ_INSERT_TAIL(&interfaces, ifp, if_list);
      TAILQ_INIT(&ifp->routers);
      TAILQ_INIT(&ifp->prefixes);
      
      ifp->autoconf = 1;
      ifp->stateful = RTADV_FLG;
      ifp->ntosend  = MAX_RTR_SOLICITATIONS;
         
      if (verbose)
	 logtrace("Read interface %s\n", ifp->name);
           
      /* 
       * Retrieve all addresses 
       */
      strcpy(ifr.ifr_name, ifp->name);
      ifr.ifr_addr.sa_family = AF_INET6;
      ifr.ifr_len6 = sizeof(buff2);
      ifr.ifr_buf6 = (caddr_t) &buff2;
      if (ioctl(u, SIOCGIFADDR, &ifr) < 0) { 
	 fprintf(stderr, "Pb getting ipv6 addresses for %s\n", ifp->name);
	 return;
      }
      itf_addr_nb = ifr.ifr_len6 / sizeof(struct in6_addr);
      buf2 = buff2;

      /*
       * Update prefix list
       */
      for (; itf_addr_nb > 0; 
	   buf2 += sizeof(struct ifreq6), itf_addr_nb --) {
	 /* Alloc prefix struct memory */
	 if (Calloc(pp, prefix) == NULL) {
	    if (verbose)
	       logerr("out of memory\n");
	    continue;
	 }
	  
	 /* Prefix length */
	 pp->plen = ((struct ifreq6 *)buf2)->preflen;
	 
	 /* IPv6 Address */
	 memcpy(&pp->addr, &(((struct ifreq6 *)buf2)->in6), 
		sizeof (struct in6_addr));
	       
	 if (verbose)
	    fprintf(stderr, "Interface %s has addr %s plen %d \n",
		    ifp->name,
		    inet_ntop(AF_INET6, &pp->addr, 
			      inet_ntop_ret, INET6_ADDRSTRLEN), pp->plen);
	    
	 /* Null Addr??? */
	 if (IS_ANYADDR6(pp->addr)) 
	    goto loop_on_error;
	 
	 /* Link-local Addr? (Does not need be advertised!) */
	 if (IS_LINKLADDR6(pp->addr)) {
	    if (IS_ANYADDR6(ifp->llocal)) {
	       COPY_ADDR6(ifp->llocal, pp->addr);	
	       printf("Interface link-local update: %s\n", 
		      inet_ntop(AF_INET6, &pp->addr,
				inet_ntop_ret, INET6_ADDRSTRLEN));
	    }
	    goto loop_on_error;
	 }
	 
	 /* ??, multicast addr. */
	 if ((pp->addr.s6_addr8[0] == 0) ||
	     (pp->addr.s6_addr8[0] == 0xff))
	    goto loop_on_error;

	 /* Interface */
	 pp->ifp = ifp;
	 
	 /* Prefix */
	 COPY_ADDR6(pp->pref, pp->addr);
	 mk_pref(pp);
	 mk_addr(pp);
#ifdef NDPD_HOST_DEBUG	    
	 printf("%s\n",
		inet_ntop(AF_INET6, &pp->addr, 
			  inet_ntop_ret, INET6_ADDRSTRLEN));
	 printf("%s\n",
		inet_ntop(AF_INET6, &pp->pref, 
			  inet_ntop_ret, INET6_ADDRSTRLEN));
#endif
	 /* Life times */
	 pp->invlife = INFLIFE;
	 pp->deplife = DEFDLIFE;
	 
	 /* Flags */
	 pp->nflags = ND_PREF_FLG_A;
	 if (pp->plen != 128)
	    pp->nflags |= ND_PREF_FLG_L;
	 pp->pflags = pp->nflags;
	 
	 TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
	 continue;

loop_on_error:
	 if (verbose)
	    logerr("Ignored interface address %s\n",
		   inet_ntop(AF_INET6, &pp->addr,
			     inet_ntop_ret, INET6_ADDRSTRLEN));
	 free(pp);	 
      }

   }

   for (ifp = interfaces.tqh_first; ifp;) {
      /*
       * Delete interface without link-local addr
       */
      if (IS_ANYADDR6(ifp->llocal)) {
	 if (verbose)
	    logtrace("delete interface %s without LL\n", ifp->name); 
	
	 ifp_tmp = ifp;
	 ifp = ifp->if_list.tqe_next;
	 TAILQ_REMOVE(&interfaces, ifp_tmp, if_list);	
	 for (pp = FIRST(ifp_tmp); pp; pp = pp->prf_list.tqe_next) {	    
	    free(pp);
	 }
	 free(ifp_tmp);
	 continue;
      }
#if 1
      dump_hex(ifp->token, LTOKEN);
      dump_hex(&ifp->llocal.s6_addr8[16-LTOKEN], LTOKEN);
#endif 
      
      /*
       * Update token based on link-local address EUI64
       */
      if (bcmp(&ifp->llocal.s6_addr8[16-LTOKEN], ifp->token, LTOKEN) != 0) {
	 if (verbose)
	    logtrace("update MAC of interface %s\n", ifp->name);
	 bcopy(&ifp->llocal.s6_addr8[16-LTOKEN], ifp->token, LTOKEN);
      }
     
      if (FIRST(ifp) == NULL) {
	 COPY_ADDR6(ifp->addr, ifp->llocal);
      } else {
	 COPY_ADDR6(ifp->addr, FIRST(ifp)->addr);
      }
    
      /*
       * Update prefixes information
       */
      for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next) {
	 update_prefix(pp);
      }
      
      /*
       * Sort interface prefixes
       */
      sort_prefix(ifp);
      ifp = ifp->if_list.tqe_next;
      /* prompt(ifp); modifies address order in kernel */
   }
}
/*--START FUNCTION--(get_main_if)---------------------------------------------

  Retrieve main interface from /proc/net6/route6
  (the one the default route is assigned to)

-----------------------------------------------------------------------------*/
void
get_main_if() {
   struct interface *ifp;
   int plen;
   unsigned long a0, a1, a2, a3;
   FILE *fp;
   char dev[6], 
        line[225],
        *eof = NULL;
   register int i;
   
   /* 
    * Open route file
    */
   fp = fopen("/proc/net6/route6", "r");
   do {
      eof = fgets(line, 255, fp);
      i = sscanf(line, "%s"
		       "\t%08lX:%08lX:%08lX:%08lX"
		       "\t%*08X:%*08X:%*08X:%*08X"
		       "\t%*04X"
		       "\t%*d"
		       "\t%*d"
		       "\t%*d"
		       "\t%d"
		       "\t%*d"
		       "\t%*d"
		       "\t%*u\n", dev, &a0, &a1, &a2, &a3, &plen);
   } while ((eof) && (a0 || a1 || a2 || a3 || plen));
   fclose(fp);
   
   if ((a0 || a1 || a2 || a3 || plen)) {
      fprintf(stderr, "can't find default IEEE interface?\n");
      exit(4); /* FIXME: was EX_OSFILE ? */
   }
   
   for (ifp = interfaces.tqh_first; ifp ; ifp = ifp->if_list.tqe_next)
      if (strcasecmp(ifp->name, dev) == 0)
	 break;
   
   main_if = ifp;
   if (verbose) {
      if (ifp)
	 logtrace("Main interface is %s\n", ifp->name);
      else
	 logtrace("No main interface?\n"); 
   }
}












