#include <string.h>
#include <stdlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include <sys/queue.h>

#include <net/if.h>
#include <net/dret/route6.h>
#include <net/dret/nd.h>
#include <net/netlink.h>

#include <linux/in.h>
#include <linux/dret/in6.h>
#include <linux/dret/ipv6.h>
#include <linux/dret/icmpv6.h>
#include <linux/un.h>
#include <linux/route.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>


#include <asm/checksum.h>

#define INET6
/* #define NDPD_HOST_DEBUG  */


#ifndef TAILQ_INSERT_BEFORE
#define	TAILQ_INSERT_BEFORE(listelm, elm, field) {			\
	(elm)->field.tqe_prev = (listelm)->field.tqe_prev;		\
	(elm)->field.tqe_next = (listelm);				\
	*(listelm)->field.tqe_prev = (elm);				\
	(listelm)->field.tqe_prev = &(elm)->field.tqe_next;		\
}
#endif

#define MAXTTL		255
#define	LTOKEN		8	/* should be variable ? */
#define WAIT_INIT	30

#define RTADV_FLG_M	0x80
#define RTADV_FLG_O	0x40
#define RTADV_FLG	(RTADV_FLG_M|RTADV_FLG_O)

#define MAX_RTR_SOLICITATIONS 		3
#define RTR_SOLICITATION_INTERVAL 	3.0
#define MAX_RTR_SOLICITATION_DELAY	1.0

#define Calloc(p, t)	(p = (struct t *)calloc(1,sizeof(struct t)))

struct prefix {
	TAILQ_ENTRY(prefix) prf_list;	/* prefix list */    
	struct interface *ifp;		/* up pointer */
	struct in6_addr pref;		/* prefix itself */
	struct in6_addr pmask;		/* prefix mask */
	struct in6_addr addr;		/* address */
	struct in6_addr amask;		/* address mask */
	struct timeout *tp;		/* timeout */
	u_int invlife;			/* invalidation lifetime */
	u_int deplife;			/* deprecation lifetime */
	u_int rendelay;			/* renew delay */
	u_char plen;			/* length */
	u_char pflags;			/* flags */
	u_char nflags;			/* new flags */
};

#define INFLIFE		0xffffffff
#define DEFDLIFE	(7 * 24 * 60 * 60)

struct router {
	TAILQ_ENTRY(router) rt_list;	/* router list */
	struct interface *ifp;		/* up pointer */
	struct sockaddr_in6 addr;	/* designated address */
	char   phy[LTOKEN-2];	        /* data-link address */
        u_char alen;
        u_char token[LTOKEN];
	struct timeout *tp;		/* last seen */
	int preference;			/* preference */
	u_short lifetime;		/* lifetime */
	u_char aflags;			/* autoconfig flags */
	u_char nflags;			/* new flags */
	u_int reach;			/* ND reachable time */
	u_int retrans;			/* ND reachable retrans time */
};

struct interface {
	TAILQ_ENTRY(interface) if_list;	/* list of interfaces */
	TAILQ_HEAD(, router) routers;	/* routers */
	TAILQ_HEAD(, prefix) prefixes;	/* prefixes */
        unsigned int index;             /* interface index */
	struct in6_addr llocal;		/* link-local address */
	struct in6_addr addr;		/* best address */
	u_char token[LTOKEN];		/* MAC address or token */
	char autoconf;			/* autoconf enabled */
	u_char stateful;		/* stateful flags */
	u_long mtu;			/* IPv6 MTU */
	short flags;			/* copy of kernel flags */
	char name[IFNAMSIZ];		/* name */
	int ntosend;			/* Rt Sol to send */
};
TAILQ_HEAD(, interface) interfaces;
struct interface *def_if, *main_if;     /* default and main if */

#define BEST(ifp)	(ifp->routers.tqh_first)
#define FIRST(ifp)      (ifp->prefixes.tqh_first)

struct timeout {
	TAILQ_ENTRY(timeout) tm_list;	/* timeout list */
	struct timeval tv;		/* when */
	void (*func) ();		/* function */
	void *arg;			/* argument */
	int expired;			/* expired */
	int refcount;			/* refcount */
};
TAILQ_HEAD(, timeout) timeouts;


struct {
	struct in6_addr	addr;		/* default source address */
	unsigned long index;		/* default interface index */
} def_ifa;


extern int pid;

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(int32_t) - 1))) : sizeof(int32_t))
#define ADVANCE(x, n) (x += sizeof(struct sockaddr_in6))

extern struct timeval now;

/* Time normalization (tv_usec less than 1 sec.) */
#define NORMTIME(tv) \
	while ((tv).tv_usec >= 1000000) { \
		(tv).tv_sec += 1; \
		(tv).tv_usec -= 1000000; \
	}

/* True if tv1 > tv2 temporally */
#define AFTER(tv1, tv2) \
	(((tv1).tv_sec > (tv2).tv_sec) || \
	 (((tv1).tv_sec == (tv2).tv_sec) && ((tv1).tv_usec >= (tv2).tv_usec)))

extern int r, s, u;			/* Socket fd: icmp, routing, udp */
extern char inet_ntop_ret[], inet_ntop_ret1[];
extern int verbose;
extern int debug;

void send_rtsol(struct interface *);
int in_cksum(u_short *, int);
void recv_icmp();
void recv_rtadv(struct interface *, struct ipv6hdr *, int);
void recv_llredirect(struct interface *, struct ipv6hdr *, int);
void recv_rtredirect(struct interface *, struct ipv6hdr *, int);
void recv_route();
void init(), set_verbose(int), dump(), finish();
void get_main_if();
void update_prefix(struct prefix *);
void prefix_deprec(struct prefix *);
void prefix_inval(struct prefix *);
void sort_prefix(struct interface *);
void add_prefix(struct prefix *);
void del_prefix(struct prefix *);
void prompt(struct interface *);
void add_prefroute(struct prefix *);
void del_prefroute(struct prefix *);
void update_router(struct router *);
void router_invalid(struct router *);
void router_unreachable(struct router *);
void sort_router(struct interface *, int);
void get_rtphy(struct router *);
void router_lost(struct sockaddr_in6 *);
void reset_default();
void add_default(int);
void del_default();
void set_default();
void set_defifa();
void add_timeout(struct timeout *);
void dump_prefix(struct prefix *);
void dump_router(struct router *);
void dump_interface(struct interface *);
void dump_timeout(struct timeout *);
void mk_addr(struct prefix *);
void mk_mask(struct in6_addr *, int);
void mk_pref(struct prefix *);
u_char get_prefixlen(struct sockaddr_in6 *); 
char *mac2ascii(caddr_t);
void initlog();
void logerr(), logtrace(), logdebug();
void logperror(char *);
void dump_hex(char *, int);











