#include "ndpd.h"

/*-----------------------------------------------------------------------
  update_prefix OK
  prefix_deprec
  prefix_inval
  sort_prefix
  add_prefix
  del_prefix OK
  prompt
  add_prefroute
  del_prefroute
  update_router
  router_invalid
  router_unreachable
  sort_router
--------------------------------------------------------------------------*/  
struct router *def_rt;
struct ifreq ifr; 
int booting = 1;


/*--START FUNCTION--(update_prefix)-------------------------------------------

 Called from init()
 Manages events concerning modified prefixes.

---------------------------------------------------------------------------*/
void
update_prefix(pp)
	register struct prefix *pp;
{
        if (verbose)
		logtrace("In update prefix\n");
	
	(void)gettimeofday(&now, NULL);

	if (pp->tp) {
		if (pp->tp->expired == 0)
			TAILQ_REMOVE(&timeouts, pp->tp, tm_list);
		if (--pp->tp->refcount <= 0)
			free(pp->tp);
		pp->tp = NULL;
	}

	pp->nflags &= ~1;
	printf("FLAGS: %X %X\n", pp->nflags, pp->pflags);
        /* auto flag turned off */
	if ((pp->pflags != pp->nflags) && (pp->pflags & ND_PREF_FLG_A))
		del_prefix(pp);
        /* onlink turned off */
	if ((pp->pflags & ND_PREF_FLG_L) && 
	    (!(pp->nflags & ND_PREF_FLG_L))) {
	        if (verbose)
			logtrace("Onlink turned off\n");
		del_prefroute(pp);
	}
        /* onlink turned on */
	if ((!(pp->pflags & ND_PREF_FLG_L)) && 
	    (pp->nflags & ND_PREF_FLG_L)) {
	        if (verbose)
			logtrace("Onlink turned on\n");
		add_prefroute(pp);
	}
        /* auto turned on */
	if ((!(pp->pflags & ND_PREF_FLG_A)) &&
	    (pp->nflags & ND_PREF_FLG_A)) {
		if (verbose)
			logtrace("auto turned on\n");
		CLR_ADDR6(pp->amask);
		if (pp->nflags & ND_PREF_FLG_L) {
			COPY_ADDR6(pp->amask, pp->pmask);
		} else {
			mk_mask(&pp->amask, 128);
		}
		add_prefix(pp);

		if (pp == FIRST(pp->ifp))
			prompt(pp->ifp);
	}
	pp->pflags = pp->nflags;
        
        /* prefix becomes invalid */
	if (pp->invlife == 0) {
		prefix_inval(pp);
		return;
	}
        /* prefix becomes deprecated */
	if (pp->deplife == 0) {
		prefix_deprec(pp);
		return;
	}
        /* deprecated lt not infinite, start timer */
	if (pp->deplife != INFLIFE) {
		if (Calloc(pp->tp, timeout) == NULL) {
			if (verbose)
				logerr("out of memory\n");
			return;
		}
		pp->tp->tv.tv_sec = now.tv_sec + pp->deplife;
		pp->tp->tv.tv_usec = now.tv_usec;
		pp->tp->func = prefix_deprec;
		pp->tp->arg = pp;
		pp->tp->refcount++;
		add_timeout(pp->tp);
		return;
	}
        /* invalidation time not infinite, start timer */
	if (pp->invlife != INFLIFE) {
		if (Calloc(pp->tp, timeout) == NULL) {
			if (verbose)
				logerr("out of memory\n");
			return;
		}
		pp->tp->tv.tv_sec = now.tv_sec + pp->invlife;
		pp->tp->tv.tv_usec = now.tv_usec;
		pp->tp->func = prefix_inval;
		pp->tp->arg = pp;
		pp->tp->refcount++;
		add_timeout(pp->tp);
		return;
	}
}

/*-------------------------------------------------------------------------

  Called from update_prefix


----------------------------------------------------------------------------*/
void
prefix_deprec(pp)
	register struct prefix *pp;
{
	struct prefix *first = FIRST(pp->ifp);

	(void)gettimeofday(&now, NULL);

	if (pp->tp) {
		if (pp->tp->expired == 0)
			TAILQ_REMOVE(&timeouts, pp->tp, tm_list);
		if (--pp->tp->refcount <= 0)
			free(pp->tp);
		pp->tp = NULL;
	}

	TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);
	TAILQ_INSERT_TAIL(&pp->ifp->prefixes, pp, prf_list);

        /* deprec. time same as invalid. time */
	if (pp->invlife == pp->deplife)
		prefix_inval(pp);
        
        /* next step is invalidation, start timer */
	if (pp->invlife != INFLIFE) {
		if (Calloc(pp->tp, timeout) == NULL) {
			if (verbose)
				logerr("out of memory\n");
			return;
		}

		pp->tp->tv.tv_sec = now.tv_sec + pp->invlife - pp->deplife;
		pp->tp->tv.tv_usec = now.tv_usec;
		pp->tp->func = prefix_inval;
		pp->tp->arg = pp;
		pp->tp->refcount++;
		add_timeout(pp->tp);
	}

	if (verbose)
		logtrace("deprecated prefix %s on %s\n",
			 inet_ntop(AF_INET6, &pp->addr,
				    inet_ntop_ret, INET6_ADDRSTRLEN),
			 pp->ifp->name);
	if (pp == first)
		sort_prefix(pp->ifp);
}
/*--END FUNCTION--()-------------------------------------------*/

/*--------------------------------------------------------------------------
  
   Called from update_prefix()
      "     "  prefix_deprec()
      "     "  recv_route() [DELADDR]
--------------------------------------------------------------------------*/
void
prefix_inval(register struct prefix *pp) {
   struct prefix *first = FIRST(pp->ifp);

   if (pp->tp) {
      if (pp->tp->expired == 0)
	 TAILQ_REMOVE(&timeouts, pp->tp, tm_list);
      if (--pp->tp->refcount <= 0)
	 free(pp->tp);
      pp->tp = NULL;
   }

   del_prefix(pp);
   TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);

   if (verbose)
      logtrace("Invalidated prefix %s on %s\n",
	       inet_ntop(AF_INET6, &pp->addr,
			 inet_ntop_ret, INET6_ADDRSTRLEN),
	       pp->ifp->name);
   if (pp == first)
      sort_prefix(pp->ifp);
   free(pp);
}
/*--END FUNCTION-----------------------------------------------------------*/

/*--START FUNCTION--()---------------------------------------------

   Sort prefixes according to their lifetimes

---------------------------------------------------------------------------*/
void
sort_prefix(ifp)
	register struct interface *ifp;
{
	register struct prefix *pp, *pp2;
	int changed = 0;
	
#define SWAP_PREFIX	\
	if (pp == FIRST(ifp)) \
		changed++; \
	TAILQ_REMOVE(&ifp->prefixes, pp2, prf_list); \
	TAILQ_INSERT_BEFORE(pp, pp2, prf_list); \
	if (pp == FIRST(ifp)) \
		changed++; \
	goto again;

		
    again:

	for (pp = FIRST(ifp); pp; pp = pp2) {
		pp2 = pp->prf_list.tqe_next;
		if (pp2 == NULL)
			break;
		if (IS_SITELADDR6(pp->addr)) {
			if (!IS_SITELADDR6(pp2->addr)) {
				SWAP_PREFIX
			}
			if ((pp->deplife < pp2->deplife) ||
			    ((pp->deplife == pp2->deplife) &&
			     (pp->invlife < pp2->invlife))) {
				SWAP_PREFIX
			}
		} else if (IS_SITELADDR6(pp2->addr)) {
			continue;
		} else if ((pp->deplife < pp2->deplife) ||
			   ((pp->deplife == pp2->deplife) &&
			    (pp->invlife < pp2->invlife))) {
			SWAP_PREFIX
		}
	}

	if (FIRST(ifp) == NULL) {
		COPY_ADDR6(ifp->llocal, ifp->addr);
		changed++;
	} else if(changed) {
		COPY_ADDR6(FIRST(ifp)->addr, ifp->addr);
	}

	if (changed == 0)
		return;
        
	prompt(ifp);
       
	if ((booting == 0) &&
	    (((pp = FIRST(ifp)) == NULL) ||
	     IS_SITELADDR6(pp->addr) ||
	     (pp->deplife == 0)))
		reset_default();
}
/*-----------------------------------------------------------------------*/


/*--START FUNCTION--()----------------------------------------------------



-------------------------------------------------------------------------*/
void
add_prefix(pp)
	register struct prefix *pp;
{
        struct ifreq6 req6;
	
	bzero((caddr_t)&ifr, sizeof(ifr));
	COPY_ADDR6(req6.in6, pp->addr);
	req6.preflen = pp->plen;
	ifr.ifr_buf6 = &req6;
        ifr.ifr_len6 = sizeof(req6);	
	strncpy(ifr.ifr_name, pp->ifp->name, IFNAMSIZ);
        ifr.ifr_addr.sa_family = AF_INET6;
	
	if (ioctl(u, SIOCSIFADDR, (caddr_t)&ifr) < 0) {
		logperror("ioctl (SIOCSIFADDR)");
		logerr("Can't add prefix %s on interface %s\n",
		       inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
				  INET6_ADDRSTRLEN),
		       pp->ifp->name);
        } else if (verbose)
	        logtrace("Added prefix %s on interface %s\n",
			 inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
				    INET6_ADDRSTRLEN),
			 pp->ifp->name);
}
/*--START FUNCTION--(del_prefix)-------------------------------------------

  Delete the prefix's address in kernel
  
  ------------------------------------------------------------------------*/
void
del_prefix(register struct prefix *pp) {
   struct ifreq6 req6;
	
   bzero((caddr_t)&ifr, sizeof(ifr));
   COPY_ADDR6(req6.in6, pp->addr);
   req6.preflen = pp->plen;
   ifr.ifr_buf6 = &req6;
   ifr.ifr_len6 = sizeof(req6);	
   strncpy(ifr.ifr_name, pp->ifp->name, IFNAMSIZ);
   ifr.ifr_addr.sa_family = AF_INET6;
	
   if (ioctl(u, SIOCDELADDR, (caddr_t)&ifr) < 0) {
      logperror("ioctl (SIOCDELADDR)");
      logerr("Can't delete prefix %s on interface %s\n",
	     inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
		       INET6_ADDRSTRLEN),
	     pp->ifp->name);
   } else if (verbose)
      logtrace("Deleted prefix %s on interface %s\n",
	       inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret, INET6_ADDRSTRLEN),
	       pp->ifp->name);			 
}

/*--START FUNCTION--(prompt)-------------------------------------------

  Should put address in front of list for the considered interface.
  (FIXME in kernel)

------------------------------------------------------------------*/
void
prompt(ifp)
	register struct interface *ifp;
{
        struct ifreq6 req6;
	
	
	bzero((caddr_t)&ifr, sizeof(ifr));
	
	strcpy(ifr.ifr_name, ifp->name);
      
	COPY_ADDR6(req6.in6, ifp->addr);
     
	ifr.ifr_buf6 = &req6;
      
        ifr.ifr_len6 = sizeof(req6);
    
        
	/*
	if (ioctl(u, SIOCFIFADDR6, (caddr_t)&ifr) < 0) {
		logperror("ioctl (SIOCFIFADDR6)");
		logerr("can't prompt prefix %s on interface %s\n",
		       inet_ntop(AF_INET6, &ifp->addr,inet_ntop_ret
				  INET6_ADDRSTRLEN),
		       ifp->name);
	} else if (verbose)
		logtrace("prompt prefix %s on interface %s\n",
			 inet_ntop(AF_INET6, &ifp->addr, inet_ntop_ret
				    INET6_ADDRSTRLEN),
			 ifp->name);

	if (ifp == def_if)
		set_defifa();
	*/
}

/*--START FUNCTION--(add_prefroute)------------------------------------------
  
  Add a route after receiving a prefix advertisement
  
  --------------------------------------------------------------------------*/
void
add_prefroute(register struct prefix *pp) {
   struct in6_rtmsg rt;

   bzero((caddr_t)&rt, sizeof(rt));
   rt.rtmsg_type = SIOCADDRT;
   rt.rtmsg_flags = RTF_UP;
   COPY_ADDR6(rt.rtmsg_dst, pp->pref);
   rt.rtmsg_prefixlen = pp->plen;
   strcpy(rt.rtmsg_device, pp->ifp->name);
	
   if (ioctl(u, SIOCADDRT, &rt) < 0) {
      if (verbose)
	 logperror("Add prefix route error\n");
      return;
   } else if (verbose)
      logtrace("Added prefix route %s %s/%d\n",
	       rt.rtmsg_device,
	       inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
			 INET6_ADDRSTRLEN),
	       pp->plen); 
}

/*--START FUNCTION--(del_prefroute)-----------------------------------------
  
  -------------------------------------------------------------------------*/
void
del_prefroute(register struct prefix * pp) {
   struct in6_rtmsg rt;

   bzero((caddr_t)&rt, sizeof(rt));
   rt.rtmsg_type = SIOCDELRT;
   rt.rtmsg_flags = RTF_UP;
   COPY_ADDR6(rt.rtmsg_dst, pp->pref);
   rt.rtmsg_prefixlen = pp->plen;
   strcpy(rt.rtmsg_device, pp->ifp->name);
	
   if (ioctl(u, SIOCDELRT, &rt) < 0) {
      if (verbose)
	 logperror("Delete prefix route error\n");
      return;
   } else if (verbose)
      logtrace("Deleted prefix route %s/%d\n",
	       inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
			 INET6_ADDRSTRLEN),
	       pp->plen);
}



void
update_router(rp)
	register struct router *rp;
{
  /*	(void)gettimeofday(&now, NULL);

	if (rp->tp) {
		if (rp->tp->expired == 0)
			TAILQ_REMOVE(&timeouts, rp->tp, tm_list);
		if (--rp->tp->refcount <= 0)
			free(rp->tp);
		rp->tp = NULL;
	}

	if (rp->aflags != rp->nflags) {
	      
#ifdef notyet
		stateful(rp->aflags, rp->nflags);
#endif
		rp->ifp->stateful = (rp->nflags & RTADV_FLG);
	}
	rp->aflags = rp->nflags;

	if (rp->phy.sdl_alen == 0)
		rp->preference = -1;
	else
		rp->preference = 0;
	if (rp->lifetime == 0) {
		router_invalid(rp);
		return;
	}

	if (rp->phy.sdl_alen &&
	    ((rp->preference > BEST(rp->ifp)->preference) ||
	     (BEST(rp->ifp)->phy.sdl_alen == 0))) {
		TAILQ_REMOVE(&rp->ifp->routers, rp, rt_list);
		TAILQ_INSERT_HEAD(&rp->ifp->routers, rp, rt_list);
		if (rp->ifp == def_if)
			set_default();
	}

	if (Calloc(rp->tp, timeout) == NULL) {
		if (verbose)
			logerr("out of memory\n");
		return;
	}

	rp->tp->tv.tv_sec = now.tv_sec + rp->lifetime;
	rp->tp->tv.tv_usec = now.tv_usec;
	rp->tp->func = router_invalid;
	rp->tp->arg = rp;
	rp->tp->refcount++;
	add_timeout(rp->tp);

	sort_router(rp->ifp, 1);
	*/
}


void
router_invalid(rp)
	register struct router *rp;
{
  /*	(void)gettimeofday(&now, NULL);

	if (rp->tp) {
		if (rp->tp->expired == 0)
			TAILQ_REMOVE(&timeouts, rp->tp, tm_list);
		if (--rp->tp->refcount <= 0)
			free(rp->tp);
		rp->tp = NULL;
	}

	if (verbose)
		logtrace("invalided router %s\n",
			 inet_ntop(AF_INET6, &rp->addr.sin6_addr, inet_ntop_ret
				    INET6_ADDRSTRLEN));
	TAILQ_REMOVE(&rp->ifp->routers, rp, rt_list);
	sort_router(rp->ifp, 1);
	free(rp);
	*/
}

void
router_unreachable(rp)
	register struct router *rp;
{
  /*	rp->phy.sdl_alen = 0;
	rp->preference = -1;
	if (verbose)
		logtrace("unreachable router %s\n",
			 inet_ntop(AF_INET6, &rp->addr.sin6_addr,inet_ntop_ret
				    INET6_ADDRSTRLEN));
	sort_router(rp->ifp, 1);
	*/
}

void
sort_router(ifp, reset)
	register struct interface *ifp;
	int reset;
{
  /*	register struct router *rp, *rp2;
	int changed = 0;

#define	SWAP_ROUTER	\
  	if (rp == BEST(ifp)) \
  		changed++; \
	TAILQ_REMOVE(&ifp->routers, rp2, rt_list); \
	TAILQ_INSERT_BEFORE(rp, rp2, rt_list); \
	if (rp == BEST(ifp)) \
		changed++; \
	goto again;

    again:

	for (rp = BEST(ifp); rp; rp = rp2) {
		rp2 = rp->rt_list.tqe_next;
		if (rp2 == NULL)
			break;
		if (rp->phy.sdl_alen == 0) {
			if (rp2->phy.sdl_alen) {
				SWAP_ROUTER
			}
		} else if (rp2->phy.sdl_alen == 0) {
			continue;
		} else if (rp->preference < rp2->preference) {
			SWAP_ROUTER
		}
	}
	if(booting) {
		reset_default();
		return;
	}

	if ((changed == 0) || (reset == 0))
		return;

	if (ifp == def_if)
		reset_default();
		*/
}

/*--START FUNCTION--(get_rtphy)-----------------------------------------------
  
  ---------------------------------------------------------------------------*/
void
get_rtphy(register struct router *rp) {
   struct sockaddr_in6 *sin, dst;
   struct sockaddr_dl *sdl;
   register int i;
   int len;

   /* 
    * Send request to kernel for redirected entry
    */
	
   /* 
    * Receive request from kernel
    */
   
   /*	      
     bcopy(LLADDR(sdl), LLADDR(&rp->phy), LTOKEN);
     default:

     if (verbose)
     logtrace("%sget physical address %s of %s\n",
     rp->phy.sdl_alen ? "" : "try to ",
     rp->phy.sdl_alen ? mac2ascii(LLADDR(&rp->phy)) : "-",
     inet_ntop(AF_INET6, &rp->addr.sin6_addr, inet_ntop_ret
     INET6_ADDRSTRLEN));

     if (rp->phy.sdl_alen) {
     sort_router(rp->ifp, 0);
     return;
     }
     */

   /*
     forceaccess:
     bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
     dst.sin6_len = sizeof(struct sockaddr_in6);
     dst.sin6_family = AF_INET6;
     COPY_ADDR6(rp->addr.sin6_addr, dst.sin6_addr);
     dst.sin6_port = htons(9);

     if (sendto(u, cp, 1, 0, (struct sockaddr *)&dst, dst.sin6_len) < 0) {
     if (verbose)
     logperror("send UDP");
     return;
     } else if (verbose)
     logtrace("send UDP probe to %s\n",
     inet_ntop(AF_INET6, &rp->addr.sin6_addr,inet_ntop_ret,
     INET6_ADDRSTRLEN));
     */
}

void
router_lost(raddr)
	register struct sockaddr_in6 *raddr;
{
  /*
	register char *cp = m_rtmsg.m_space;
	struct sockaddr_in6 *sin;
	int len;

	bzero((caddr_t)&m_rtmsg, sizeof(m_rtmsg));
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_type = RTM_DELETE;
	rtm.rtm_flags = RTF_UP|RTF_GATEWAY|RTF_HOST|RTF_STATIC;
	rtm.rtm_addrs = RTA_DST;
	rtm.rtm_pid = pid;
	rtm.rtm_seq = ++rtm_seq;

	sin = (struct sockaddr_in6 *)cp;
	sin->sin6_family = AF_INET6;
	sin->sin6_len = sizeof(struct sockaddr_in6);
	COPY_ADDR6(raddr->sin6_addr, sin->sin6_addr);
	ADVANCE(cp, sin);

	rtm.rtm_msglen = len = cp - (char *)&m_rtmsg;
	cp = (caddr_t)&m_rtmsg;

       	if (write(r, cp, len) < len) {
		if (verbose)
			logperror("write (route message)");
		return;
	} else if (verbose)
		logtrace("lost router %s\n",
			 inet_ntop(AF_INET6, &raddr->sin6_addr, inet_ntop_ret,
				    INET6_ADDRSTRLEN));
				    */
}

/*--START FUNCTION--(reset_default)-------------------------------------------
  
  Re-elect default interface according to # of criteria
  
  -------------------------------------------------------------------------*/
void
reset_default() 
{
   register struct interface *ifp, 
                   *best_if = main_if;
   register struct router *rp;
   register struct prefix *pp;
   struct timeout *tp;
   int goodies = 0;                /* the greater, the better the interface */

   for (ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next) {
      /*
       * First IF up
       */
      if (ifp->flags & IFF_UP) {
	 if (goodies == 0) {
	    best_if = ifp;
	    goodies++;
	 }
      } else 
	 continue;

      /* 
       * Knows about a router 
       */
      if ((rp = BEST(ifp))) {
	 if (goodies == 1) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
	
      
      for (rp = BEST(ifp); rp; rp = rp->rt_list.tqe_next)
	 if (rp->alen == 0)
	    get_rtphy(rp);    /* FIXME: not implemented
			       * Get rt link-layer address from kernel
			       */
      
      /* 
       *  
       */
      if ((rp = BEST(ifp))->alen) {
	 if (goodies == 2) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
      
      /* 
       * Router pref >= 0 
       */
      if (rp->preference >= 0) {
	 if (goodies == 3) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
      
      /* 
       * Router has a prefix ? 
       */
      if ((pp = FIRST(ifp))) {
	 if (goodies == 4) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
      
      /* 
       * Prefix is non deprecated! 
       */
      if (pp->deplife > 0) {
	 if (goodies == 5) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
      
      /* 
       * Not site local !!! 
       */
      if (!IS_SITELADDR6(pp->addr)) {
	 if (goodies == 6) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
      
      /* 
       * & BINGO!!!! 
       */
      if (ifp == def_if) {
	 if (goodies == 7) {
	    best_if = ifp;
	    goodies++;
	 }
      } else
	 continue;
      
      /* 
       * WOW !!! 
       */
      if (ifp == main_if) {
	 best_if = ifp;
	 goodies = 9;
	 break;
      }
   }
   
   if (booting) {
      if (verbose)
	 logtrace("First call to reset_default\n");
      booting = 0;
      /*        	if (!debug) {
		        debug++;
        	        initlog(); 
			} */
   }
   
   if (verbose)
      logtrace("Goodies = %d\n", goodies);

   if (goodies <= 2)
      add_default(0);
   else 
      if (def_if != best_if) {
	 def_if = best_if;
	 set_default();
	 set_defifa();
      }
   
   if (goodies >= 7)
      return;

   (void)gettimeofday(&now, NULL);

   if (Calloc(tp, timeout) == NULL) {
      if (verbose)
	 logerr("out of memory\n");
      return;
   }
   tp->tv.tv_sec = now.tv_sec + (goodies <= 2 ? 1 : 10);
   tp->tv.tv_usec = now.tv_usec;
   tp->func = reset_default;
   add_timeout(tp);	
}


static struct interface *prev_if;

/*--START FUNCTION--(add_default)--------------------------------------------
  
  which = 0 : Link local route. (from finish || reset_default)
  which = 1 : From set_default, 'coz def_rt == NULL
  -------------------------------------------------------------------------*/
void
add_default(int which) {
   register struct interface *out_if = NULL;
   register struct sockaddr_in6 *sin;
   struct in6_rtmsg rt;
   int len;

   if (which == 0) {
      out_if = (def_if ? def_if : main_if);
      if (!out_if || out_if == prev_if) {
	 if (verbose)
	    logtrace("!out_if || out_if == prev_if\n");
	 return;
      }
      prev_if = out_if;
   }
   else
      prev_if = NULL;
	

   if (which == 0) {
      if (verbose)
	 logtrace("Deleting default route\n");
      del_default();

      /*
       * Set new default interface:
       */
      if (verbose)
	 logtrace("Adding new default route on %s\n", out_if->name);
      bzero((caddr_t)&rt, sizeof(rt));
      rt.rtmsg_type = SIOCADDRT;
      rt.rtmsg_flags = RTF_UP;
      rt.rtmsg_prefixlen = 0;
      
      /*
       * Gateway address
       */
      if (BEST(out_if)) {
	 COPY_ADDR6(rt.rtmsg_gateway, (BEST(out_if))->addr.sin6_addr);   
	 rt.rtmsg_flags |= RTF_GATEWAY;
      }
      strcpy(rt.rtmsg_device, out_if->name);
	
      if (ioctl(u, SIOCADDRT, &rt) < 0) {
	 if (verbose)
	    logperror("Add default route error");
	 return;
      } 
      if (verbose && (BEST(out_if)))
	 logtrace("Added default route %s gw %s\n",
		  rt.rtmsg_device,
		  inet_ntop(AF_INET6, &((BEST(out_if))->addr.sin6_addr),
		  inet_ntop_ret, INET6_ADDRSTRLEN)); 
   }
}

void
del_default() {
   struct in6_rtmsg rt;

   bzero((caddr_t)&rt, sizeof(rt));
   rt.rtmsg_type = SIOCDELRT;
	
   if (ioctl(u, SIOCDELRT, &rt) < 0) {
      if (verbose)
	 logperror("Del default route error\n");
      return;
   } else if (verbose)
      logtrace("Deleted default route\n");
}



void
set_default()
{
   register struct router *rp = BEST(def_if);
   struct in6_rtmsg rt;
   struct sockaddr_in6 *sin;
   int len;

   if (rp == NULL) {
      logerr("No default router?\n");
      abort();
   }

   del_default();

   /*
    * Set new default interface:
    */
   bzero((caddr_t)&rt, sizeof(rt));
   rt.rtmsg_type = SIOCADDRT;
   rt.rtmsg_flags = RTF_UP|RTF_GATEWAY; 
   /* rt.rtmsg_mtu = def_if->mtu; */
   COPY_ADDR6(rt.rtmsg_gateway, rp->addr.sin6_addr); 
   rt.rtmsg_prefixlen = 0;
   strcpy(rt.rtmsg_device, def_if->name);
	
   if (ioctl(u, SIOCADDRT, &rt) < 0) {
      if (verbose)
	 logperror("Add default route error\n");
      return;
   } else if (verbose)
      logtrace("Set default via %s link#%d\n",
	       inet_ntop(AF_INET6, &rp->addr.sin6_addr,inet_ntop_ret,
			 INET6_ADDRSTRLEN),
	       rp->ifp->index);

   def_rt = rp;
}



void
set_defifa()
{
   register struct interface *ifp = def_if;
   struct sockaddr_in6 *sin;
   int len;
	
   if (ifp == NULL) {
      logerr("no default interface address?\n");
      abort();
   }

   /*
    * Up to date?
    */
   if ((ifp->index == def_ifa.index) &&
       SAME_ADDR6(ifp->addr, def_ifa.addr))
      return;

   /*
    * Default router?
    */
   if (def_rt == NULL) {
      logtrace("postpone set_defifa\n");
      return;
   }
   
   /*
    * Update default source address
    */
   def_ifa.index = ifp->index;
   COPY_ADDR6(def_ifa.addr, ifp->addr);

   /*
    * Inform kernel
    */
   /* FIXME */
}









