#include "ndpd.h"
#if 1
void dump_hex(char *buf, int nb)
{
   int i = 0, n = 0;

   for (; nb > 0; nb--) {
      printf("%02hx ", (unsigned char) buf[i]);
      i++;
      n++;
      if (n == 16) {
	 printf("\n");
	 n = 0;
      }
   }
   if (n != 0) {
      printf("\n");
   }
}				/* dump_hex */
#endif
/*-------------------------------------------------------------------------
  Inserts a timeout in timeout list
  -------------------------------------------------------------------------*/
void
add_timeout(tp)
	register struct timeout *tp;
{
	register struct timeout *tp2;

	if (timeouts.tqh_first == NULL) {
		TAILQ_INSERT_HEAD(&timeouts, tp, tm_list);
		return;
	}

	for (tp2 = timeouts.tqh_first;
	     tp2 && AFTER(tp->tv, tp2->tv);
	     tp2 = tp2->tm_list.tqe_next)
		continue;

	if (tp2 == NULL) {
		TAILQ_INSERT_TAIL(&timeouts, tp, tm_list);
		return;
	}

	TAILQ_INSERT_BEFORE(tp2, tp, tm_list);
}



void
dump_prefix(pp)
	register struct prefix *pp;
{
	logdebug("prefix (%lx):\n", (u_long)pp);
	logdebug("\tinterface=%s\n", pp->ifp->name);
	logdebug("\tprefix=%s ",
		 inet_ntop(AF_INET6, &pp->pref,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	logdebug("with len=%d and mask=%s\n",
		 pp->plen,
		 inet_ntop(AF_INET6, &pp->pmask,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	logdebug("\tif address=%s ",
		 inet_ntop(AF_INET6, &pp->addr,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	logdebug("with mask=%s\n",
		 inet_ntop(AF_INET6, &pp->amask,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	logdebug("\ttimeout=%lx\n", (u_long)(pp->tp));
	logdebug("\trenew=%u\n", pp->rendelay);
	if (pp->deplife == INFLIFE)
		logdebug("\tdeprecation=infinity\n");
	else
		logdebug("\tdeprecation=%u\n", pp->deplife);
	if ((pp->deplife == INFLIFE) || (pp->invlife == INFLIFE))
		logdebug("\tinvalidation=infinity\n");
	else
		logdebug("\tinvalidation=%u\n", pp->invlife);
	logdebug("\tflags=%x (next=%x)\n", pp->pflags, pp->nflags);
}

void
dump_router(rp)
	register struct router *rp;
{
	logdebug("router (%lx):\n", (u_long)rp);
	logdebug("\tinterface=%s\n", rp->ifp->name);
	logdebug("\taddress=%s\n",
		 inet_ntop(AF_INET6, &rp->addr.sin6_addr,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	/*	logdebug("\tethernet=(link#%d) %s\n",
		 rp->phy.sdl_index,
		 rp->phy.sdl_alen ? mac2ascii(LLADDR(&rp->phy)) : "unknown");*/
	logdebug("\ttimeout=%lx\n", (u_long)(rp->tp));
	logdebug("\tpreference=%d\n", rp->preference);
	logdebug("\tlifetime=%u\n", rp->lifetime);
	logdebug("\tflags=%x (next=%x)\n", rp->aflags, rp->nflags);
	logdebug("\treachable time=%u\n", rp->reach);
	logdebug("\ttreachable retrans timers=%u\n", rp->retrans);
}

void
dump_interface(ifp)
	register struct interface *ifp;
{
	register struct prefix *pp;
	register struct router *rp;

	logdebug("interface (%lx):\n", (u_long)ifp);
	logdebug("\tname=%s\n", ifp->name);
	logdebug("\tindex=%d\n", ifp->index);
	logdebug("\tlink-local address=%s\n",
		 inet_ntop(AF_INET6, &ifp->llocal,
			    inet_ntop_ret, INET6_ADDRSTRLEN));
	logdebug("\tbest address=%s\n",
		 inet_ntop(AF_INET6, &ifp->addr,
			     inet_ntop_ret, INET6_ADDRSTRLEN));
	/*logdebug("\ttoken=%s\n", mac2ascii(ifp->token));*/
	logdebug("\tautoconf=%x\n", ifp->autoconf);
	logdebug("\tstateful=%x\n", ifp->stateful);
	logdebug("\tMTU=%lu\n", ifp->mtu);
	logdebug("\tflags=%x\n", ifp->flags);
	logdebug("\tRtSol to send=%d\n", ifp->ntosend);

	logdebug("prefix list:\n");
	for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next)
		dump_prefix(pp);

	logdebug("router list:\n");
	for (rp = ifp->routers.tqh_first; rp; rp = rp->rt_list.tqe_next)
		dump_router(rp);
}

void
dump_timeout(tp)
	register struct timeout *tp;
{
	struct timeval delta;
	int sign = 0;

	logdebug("timeout (%lx):\n", (u_long)tp);
	if (AFTER(tp->tv, now)) {
		sign = +1;
		delta.tv_sec = tp->tv.tv_sec - now.tv_sec - 1;
		delta.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
		NORMTIME(delta);
	} else {
		sign = -1;
		delta.tv_sec =  now.tv_sec - tp->tv.tv_sec - 1;
		delta.tv_usec = now.tv_usec + 1000000 - tp->tv.tv_usec;
		NORMTIME(delta);
	}
	logdebug("\texpire=%+ld.%06lu\n", sign * delta.tv_sec, delta.tv_usec);
	logdebug("\texpired=%d\n", tp->expired);
	logdebug("\trefcount=%d\n", tp->refcount);
	logdebug("\tfunction=%lx\n", (u_long)(tp->func));
	logdebug("\targument=%lx\n", (u_long)(tp->arg));
}

/*---------------------------------------------------------------------------
  No amask, ND_PREF_FLG_L  => amask = pmask
       "    !ND_PREF_FLG_L => amask of 128.
  addr = pref
  addr = addr ANDed amask
  addr = addr + token
 ---------------------------------------------------------------------------*/
void
mk_addr(pp)
	register struct prefix *pp;
{
	register int i;

	if (IS_ANYADDR6(pp->amask)) {
		if (pp->nflags & ND_PREF_FLG_L) {
			COPY_ADDR6(pp->amask, pp->pmask);
		} else {
			mk_mask(&pp->amask, 128);
		}
	}
	COPY_ADDR6(pp->addr, pp->pref);
	for (i = 0; i < 16; i++)
		pp->addr.s6_addr8[i] &= pp->amask.s6_addr8[i];
	bcopy((caddr_t)pp->ifp->token, &pp->addr.s6_addr8[16-LTOKEN], LTOKEN);
	/*dump_hex(&pp->addr.s6_addr8[16-LTOKEN], LTOKEN);*/
}

/*---------------------------------------------------------------------------
  Paddr of prefix structure ANDed with pmask of lenth plen
  --------------------------------------------------------------------------*/
void
mk_pref(pp)
	register struct prefix *pp;
{
	register int i;

	mk_mask(&pp->pmask, pp->plen);
	for (i = 0; i < 16; i++)
		pp->pref.s6_addr8[i] &= pp->pmask.s6_addr8[i];
}

/*--------------------------------------------------------------------------
  Generate mask from prefix length
  -------------------------------------------------------------------------*/
void
mk_mask(addr, len)
	register struct in6_addr *addr;
	register int len;
{
	register u_char *cp;

	CLR_ADDR6(*addr);
	cp = (u_char *)addr;
	for (cp = (u_char *)addr; len > 7; len -= 8)
		*cp++ = 0xff;
	if(len)
		*cp = 0xff << (8 - len);
}


u_char
get_prefixlen(sa)
   struct sockaddr_in6 *sa;
{
        register u_char *cp;
        register int byte, bit, len;

        if (sa->sin6_family != AF_INET6)
		return(0);

        cp = (u_char *)sa;

	/* port & flowinfo fields set to 0 */
        for (byte = 1, cp += 1; byte < 7; byte++, cp++)
        	if (*cp)
        		return(0);

        len = 0;
        for (byte = 0; byte < 16; byte++, cp++, len += 8)
	        if (*cp != 0xff)
	        	break;
        if (len == 128)
	        return((u_char)128);

        for (bit = 7; bit > 0; bit--, len++)
	        if (!(*cp & (1 << bit)))
	        	break;
	
	/* Case where 11111101 */
        for (; bit > 0; bit--)
        	if (*cp & (1 << bit))
        		return(0);

	byte++; cp++;
        for (; byte < 16; byte++, cp++)
        	if (*cp)
        		return(0);
        return((u_char)len);
}

#if (LTOKEN == 6)
char *
mac2ascii(addr)
	register caddr_t addr;
{
	static char buf[64];

	sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x",
		(u_char) addr[0], (u_char) addr[1], (u_char) addr[2],
		(u_char) addr[3], (u_char) addr[4], (u_char) addr[5]);
	return(buf);
}
#endif

static logging = 0;

void
initlog()
{
	logging++;
	openlog("ndpd", LOG_PID | LOG_CONS, LOG_DAEMON);
}

/* VARARGS1 */
void
logerr(fmt, a,b,c,d,e,f,g,h)
	char *fmt;
{
	if (logging)
		syslog(LOG_ERR, fmt, a,b,c,d,e,f,g,h);
	else
		(void) fprintf(stderr, fmt, a,b,c,d,e,f,g,h);
}

/* VARARGS1 */
void
logtrace(fmt, a,b,c,d,e,f,g,h)
	char *fmt;
{
	if (logging)
		syslog(LOG_INFO, fmt, a,b,c,d,e,f,g,h);
	else
		(void) fprintf(stdout, fmt, a,b,c,d,e,f,g,h);
}

/* VARARGS1 */
void
logdebug(fmt, a,b,c,d,e,f,g,h)
	char *fmt;
{
	if (logging)
		syslog(LOG_DEBUG, fmt, a,b,c,d,e,f,g,h);
	else
		(void) fprintf(stdout, fmt, a,b,c,d,e,f,g,h);
}

void
logperror(str)
	char *str;
{
	if (logging)
		syslog(LOG_ERR, "%s: %s\n", str, sys_errlist[errno]);
	else
		(void) fprintf(stderr, "%s: %s\n", str, sys_errlist[errno]);
}
