#include "ndpd.h"

#if 1
static void dump_hex(char *buf, int nb)
{
   int i = 0, n = 0;

   for (; nb > 0; nb--) {
      printf("%02hx ", (unsigned char) buf[i]);
      i++;
      n++;
      if (n == 16) {
	 printf("\n");
	 n = 0;
      }
   }
   if (n != 0) {
      printf("\n");
   }
}				/* dump_hex */
#endif

struct {
        struct cmsghdr c_cm;
        u_char c_space[64];
} c_cmsg;
#define cm      c_cmsg.c_cm

#define	MAXPACKET	8192	/* max packet size */
char packet[MAXPACKET];

/* Link-layer address option */
#define LLA(x)	((struct nd_opt *)(x))

/* Prefix information option */
#define PREF(x)	((struct pfx_opt *)(x))

/* MTU option */
#define MTU(x)	((struct mtu_opt *)(x))

/* shoud be a parameter ???? */
#define ND_PREF_FLAGS (ND_PREF_FLG_L|ND_PREF_FLG_A)


/*--START FUNCTION--(send_rtadv)-------------------------------------------
   Send an ICMPv6 Router Advertisment packet on ifp. 
   always use multicast 
 ------------------------------------------------------------------------*/
static void
send_rtadv(struct interface *ifp) {
   register struct ipv6hdr *ip;              /* header IPv6 */
   register struct r_adv *icp;               /* router adv. header */
   register char *hp;
   /* register struct nd_opt *llaopt;
      register struct pfx_opt *popt; */
   register struct prefix *pp;               /* prefix information */
   struct sockaddr_in6 dst;                  /* socket address */
   int icmp_len,                             /* ICMPv6 packet length */
       total_len,                            /* IPv6 packet length */
       cc;
   unsigned int csum;                        /* ICMPv6 checksum */
   long delay;
   struct timeout *tp;                       /* timeout structure */
   int dorouter = routeradv /*&& is_router()*/;

   /*
    * Interface'snt up!
    */
   if ((ifp->flags & IFF_UP) == 0)
      return;
   
   /*
    * Destination address
    */
   bzero((caddr_t)&dst, sizeof(struct sockaddr_in6));
   dst.sin6_family = AF_INET6;
   dst.sin6_addr.s6_addr8[0] = 0xff;
   dst.sin6_addr.s6_addr8[1] = MADDR6_SCP_LINK;
   dst.sin6_addr.s6_addr8[15] = MADDR6_ALLNODES;

   if (verbose) {
      logtrace("Sending router advertisement for %s on %s\n",
	       inet_ntop(AF_INET6, &ifp->llocal, 
			 inet_ntop_ret, INET6_ADDRSTRLEN),
	       ifp->name);	   
   }
   
   pp = FIRST(ifp);
	
more:
   /*
    * Initialize packet memory area
    */
   ip = (struct ipv6hdr *)packet;
   icmp_len = sizeof(struct r_adv);
   total_len = icmp_len + sizeof(struct ipv6hdr);
   bzero(packet, total_len);
	
   /*
    * IPv6 header
    */
   ip->next_hdr = IPPROTO_ICMPV6;	
   COPY_ADDR6(ip->saddr, ifp->llocal);
   COPY_ADDR6(ip->daddr, dst.sin6_addr);

   /* 
    *ICMPv6 header 
    */	
   icp = (struct r_adv *)(ip + 1);
   icp->hdr.type = ND_ROUTER_ADVERTISEMENT;
   icp->hop_limit = MaxHop;
   icp->flags.managed =  (Managed & 0x80 ? 1:0);
   icp->flags.other   =  (Managed & 0x40 ? 1:0);
   if (dorouter)
      icp->lifetime = 0xffff;	/* parameter ??? */
   icp->reachable_t = htonl(Reach);
   icp->retrans_t = htonl(Retrans);
	
   /* 
    * Link-layer address option 
    */
   LLA(hp) = LLA((char *)(icp + 1));
   if (dorouter) {
      bzero(LLA(hp), sizeof(*LLA(hp)));
      LLA(hp)->type = ND_OPT_SLLA;
      LLA(hp)->len = 1;
      bcopy(ifp->token, LLA(hp)->lla, 3);    /* FIXME !!! */
      bcopy(ifp->token + 5, LLA(hp)->lla + 3, 3);
      /* Pb here with MAX_ADDR_LEN */
      hp  +=  LLA(hp)->len << 3;
      icmp_len += 2 + (LTOKEN - 2);
   }
   
   /* 
    * Prefix information option 
    */
   if (prefixadv)
      for (; pp; pp = pp->prf_list.tqe_next) {
	 if (icmp_len + sizeof(*PREF(hp)) + sizeof(*ip) > IPV6_MTU)
	    break;
	 bzero(PREF(hp), sizeof(*PREF(hp)));
	 PREF(hp)->type = ND_OPT_PREFIX_INFO;
	 PREF(hp)->len  = 4;
	 PREF(hp)->plen = 128-8*LTOKEN;
	 PREF(hp)->flags.onlink = PREF(hp)->flags.auto1 = 1;/* parameter ??? */
	 PREF(hp)->valid_lt = INFLIFE;		/* parameter ??? */
	 PREF(hp)->pref_lt  = INFLIFE;		/* parameter ??? */
	 COPY_ADDR6(PREF(hp)->prefix, pp->pref);
	 if (verbose) {
	    logtrace("****************************\n"
		     "Sending pref. advertisement:\n"
		     "****************************\n"
		     "type     : \t%d\n"
		     "length   : \t%d\n"
		     "plen     : \t%d\n"
		     "valid_lt : \t%d\n"
		     "pref_lt  : \t%d\n"
		     "addr     : \t%s\n",
		     PREF(hp)->type,PREF(hp)->len,PREF(hp)->plen,
		     PREF(hp)->valid_lt,PREF(hp)->pref_lt,
		     inet_ntop(AF_INET6, &pp->pref, inet_ntop_ret, 
			       INET6_ADDRSTRLEN));
	 }
	 (PREF(hp))++;
	 icmp_len += sizeof(*PREF(hp));
      }

   /*
    * MTU option
    */
   
   /* 
    * Compute ICMPv6 checksum here 
    */
   csum = csum_partial((char *)icp, icmp_len, 0);
   icp->hdr.csum = csum_ipv6_magic(&ip->saddr, &ip->daddr, icmp_len, 
				   IPPROTO_ICMPV6, csum);

   /* 
    * Finish IPv6 header 
    */
   ip->payload  = htons(icmp_len);
   ip->vrprflow = ICMP6_ND_PRIORITY | IPV6_VERSION;
   ip->hop_limit = ICMP6_ND_HOPS;
	
   /*
    * Send it away!
    */
   if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
		  &ifp->index,        
		  sizeof(ifp->index)) < 0) {
      logperror("setsockopt (IPV6_MULTICAST_IF)");
      return;
   }
   
   dump_hex((char *)ip, 40);
   
   icmp_len += sizeof(struct ipv6hdr);
   cc = sendto(s, packet, icmp_len, 0,
	       (struct sockaddr *)&dst, sizeof (struct sockaddr_in6));
   if (cc < 0 || cc != icmp_len)  {
      logerr("wrote %d chars, ret=%d\n", icmp_len, cc);
      if (cc < 0) {
	 logperror("sendto");
	 return;
      }
   } 
   
   if (verbose) {
      logtrace("******* in IPV6 *******\n"
	       "vrprflow :\t%8X\n"
	       "payload  :\t%d\n"
	       "hop_limit:\t%d\n"
	       "nxt_hdr  :\t%d\n"
	       "saddr    :\t%s\n"
	       "daddr    :\t%s\n",
	       ip->vrprflow, ntohs(ip->payload), ip->hop_limit,
	       ip->next_hdr,
	       inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret, 
			 INET6_ADDRSTRLEN), 
	       inet_ntop(AF_INET6, &ip->daddr, inet_ntop_ret1, 
			 INET6_ADDRSTRLEN));
      logtrace("******* in ICMPV6 *******\n"
	       "type     :\t%d\n"
	       "hop_limit:\t%d\n"
	       "retrans_t:\t%d\n"
	       "reach_t  :\t%d\n"
	       "managed  :\t%d\n"
	       "other    :\t%d\n",
	       icp->hdr.type, icp->hop_limit, ntohl(icp->retrans_t),
	       ntohl(icp->reachable_t), icp->flags.managed, 
	       icp->flags.other);
   }
   logtrace("Destination address: %s\n", inet_ntop(AF_INET6, &ip->daddr, 
					     inet_ntop_ret, INET6_ADDRSTRLEN));
   
   /*
    * Not all prefixes have benn transmitted in previous adv.
    */
   if (prefixadv && pp)
      goto more;
   
   /*
    * Set timer for retransmission time
    */
   if (Calloc(tp, timeout) == NULL) {
      logerr("out of memory\n");
      abort();
   }
   
   (void) gettimeofday(&now, NULL);
   ifp->last = now;
   
   delay = MinRtrAdvInterval +
           random() % (MaxRtrAdvInterval - MinRtrAdvInterval);
   
   if (ifp->ntosend < MAX_INITIAL_RTR_ADVERTISEMENTS - 1) {
      ifp->ntosend++;
      if (delay > MAX_INITIAL_RTR_ADVERT_INTERVAL)
	 delay = MAX_INITIAL_RTR_ADVERT_INTERVAL;
   }
   
   tp->tv.tv_sec = now.tv_sec + TO_SEC(delay);
   tp->tv.tv_usec = now.tv_usec + TO_USEC(delay);
   NORMTIME(tp->tv);
   tp->func = send_rtadv;
   tp->arg = ifp;              
   ifp->tp = tp;
   add_timeout(tp);
}
/*--END FUNCTION--(send_rtadv)--------------------------------------------*/

/*--START_FUNCTION--(start_rtadv)-------------------------------------------
  
  ------------------------------------------------------------------------*/
void
start_rtadv(struct interface *ifp) {
   /*
    * Check timeout list
    */
   if (ifp->tp) {
      if (ifp->tp->expired == 0)
	 TAILQ_REMOVE(&timeouts, ifp->tp, tm_list);
      if (--ifp->tp->refcount <= 0) 
	 free(ifp->tp);
      ifp->tp = NULL;
   }       
   ifp->ntosend = 0;
   /*
    * Send router advertisement
    */
   send_rtadv(ifp);
}
/*--END FUNCTION--(start_rtadv)------------------------------------------*/

void
sol_rtadv(struct interface *ifp) {
   struct timeval tv;
   long delay;
   struct timeout *tp;

   (void) gettimeofday(&now, NULL);
   tv.tv_sec = ifp->last.tv_sec + TO_SEC(MIN_DELAY_BETWEEN_RAS);
   tv.tv_usec = ifp->last.tv_usec + TO_USEC(MIN_DELAY_BETWEEN_RAS);
   NORMTIME(tv);
   if (AFTER(tv, now))
      tv = now;
   delay = random()%MAX_RA_DELAY_TIME;
   tv.tv_sec = tv.tv_sec + TO_SEC(delay);
   tv.tv_usec = tv.tv_usec + TO_USEC(delay);

   NORMTIME(tv);
   if (ifp->tp) {
      if (AFTER(tv, ifp->tp->tv))
	 return;
      if (ifp->tp->expired == 0)
	 TAILQ_REMOVE(&timeouts, ifp->tp, tm_list);
      if (--ifp->tp->refcount <= 0) 
	 free(ifp->tp);
      ifp->tp = NULL;
   }       
   if (Calloc(tp, timeout) == NULL) {
      logerr("out of memory\n");
      abort();
   }
   tp->func = send_rtadv;
   tp->arg = ifp;              
   ifp->tp = tp;
   add_timeout(tp);
}

void
stop_rtadv(ifp)
	struct interface *ifp;
{
	/* TO DO . IS IT NEEDED ???? */
}

/*--START FUNCTION--(set_router)--------------------------------------------
  
  ------------------------------------------------------------------------*/
void
set_router(struct interface *ifp) {
   struct ipv6_mreq val;

   /* 
    * Link local all routers multicast address 
    */
   bzero(&val.ipv6mr_multiaddr, sizeof (val.ipv6mr_multiaddr));
   val.ipv6mr_multiaddr.s6_addr8[0] = 0xff;
   val.ipv6mr_multiaddr.s6_addr8[1] = MADDR6_SCP_LINK;
   val.ipv6mr_multiaddr.s6_addr8[15] = MADDR6_ALLROUTERS;
   
   /* 
    * Interface IPv6 address (FIXME: check last API!!!)
    */
   val.ipv6mr_interface = ifp->index;
   if (setsockopt(s, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
		  &val, sizeof(val)) < 0) {
      logerr("setsockopt (IPV6_ADD_MEMBERSHIP) on %s (intf #%d)\n %s\n", 
	     ifp->name,
	     ifp->index,
	     inet_ntop(AF_INET6, &ifp->addr,
		       inet_ntop_ret, INET6_ADDRSTRLEN));
      logperror("setsockopt (IPV6_ADD_MEMBERSHIP)");
      exit(3);
   }
}
/*--END_FUNCTION----------------------------------------------------------*/


/* 
 * Internet checksum 
 */
int
in_cksum(u_short *addr, int len) {
   register int nleft = len;
   register u_short *w = addr;
   register u_short answer;
   u_short odd_byte = 0;
   register int sum = 0;

   /*
    *  Our algorithm is simple, using a 32 bit accumulator (sum),
    *  we add sequential 16 bit words to it, and at the end, fold
    *  back all the carry bits from the top 16 bits into the lower
    *  16 bits.
    */
   while (nleft > 1)  {
      sum += *w++;
      nleft -= 2;
   }

   /* mop up an odd byte, if necessary */
   if (nleft == 1) {
      *(u_char *)(&odd_byte) = *(u_char *)w;
      sum += odd_byte;
   }

   /*
    * add back carry outs from top 16 bits to low 16 bits
    */
   sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
   sum += (sum >> 16);			/* add carry */
   answer = ~sum;				/* truncate to 16 bits */
   return (answer);
}

/*--START FUNCTION--(recv_icmp)---------------------------------------------
  
  Receive an ICMPv6 (Router Sollicitation) packet 
  -----------------------------------------------------------------------*/
void
recv_icmp() {
   register struct ipv6hdr *ip;           /* IPv6 header */
   register struct icmpv6_hdr *icp;       /* ICMPv6 header */
   struct interface *ifp;                 /* Interface structure */
   struct msghdr msg;                     /* Message header (from kernel) */
   struct iovec iov;                      /* Iovec in message */
   int cc;                                /* recvmsg return value */

   /*
    * Initializes message structure
    */
   bzero((caddr_t)&msg, sizeof(struct msghdr));
   bzero((caddr_t)&iov, sizeof(struct iovec));
   msg.msg_iov = &iov;
   msg.msg_iov->iov_base = packet;
   msg.msg_iov->iov_len = MAXPACKET;
   msg.msg_iovlen = 1;
   msg.msg_control = (caddr_t)&c_cmsg;
   msg.msg_controllen = sizeof(c_cmsg);

   /*
    * Receive message
    */
   if ((cc = recvmsg(s, &msg, 0)) < 0) {
      if (verbose)
	 logperror("read (ICMPv6)");
      return;
   }
   
   /* error on reception */
   if (cc == 0) {
      if (verbose)
	 logerr("read ICMPv6 return 0\n");
      return;
   }
   
   /* message truncated */
   if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC)) {
      if (verbose)
	 logerr("read ICMPv6 truncated\n");
      return;
   }
   
   /* control error */
   if ((msg.msg_controllen == 0) ||
       (cm.cmsg_len<(sizeof(struct cmsghdr)+sizeof(struct in6_pktinfo)))||
       (cm.cmsg_level != IPPROTO_IPV6) ||
       (cm.cmsg_type != IPV6_PKTINFO)) {
      if (verbose)
	 logerr("read ICMPv6 doesn't give interface %d %d %d %d %d\n",
		msg.msg_controllen,
		IPPROTO_IPV6, cm.cmsg_level,
		IPV6_PKTINFO, cm.cmsg_type
		);
      return;
   }

   /*
    * Look for concerned interface
    */
   for (ifp = interfaces.tqh_first; ifp; ifp = ifp->if_list.tqe_next)
      if (ifp->index ==
	  ((struct in6_pktinfo *)cm.cmsg_data)->ipi6_ifindex) 
	 break;
   if (ifp == NULL) {
      if(verbose)
	 logerr("No corresponding interface handled, %d\n",
		((struct in6_pktinfo *)cm.cmsg_data)->ipi6_ifindex);
      return;
   }	

   /* message too short */
   ip = (struct ipv6hdr *)packet;
   if (cc < sizeof(*ip) + sizeof(struct icmpv6_hdr)) {
      if (verbose)
	 logerr("ICMPv6: too short (%d)\n", cc);
      return;
   }
   
   /* non saddr address */
   if (!IS_LINKLADDR6(ip->saddr)) {
      if (verbose)
	 logerr("Source address is not linklocal...\n");
      return;
   }
   
   /* wrong hop limit */
   if (ip->hop_limit != ICMP6_ND_HOPS) {
      if (verbose)
	 logerr("no HOPS (%d)\n", ICMP6_ND_HOPS);
      return;
   }
   
   cc -= sizeof(struct ipv6hdr);
   icp = (struct icmpv6_hdr *)(ip + 1);

   /*
    * Receive router advertisement
    */
   if (icp->type == ND_ROUTER_ADVERTISEMENT)
      recv_rtadv(ifp, ip, cc);
   
   /*
    * Receive router sollicitation
    */
   else if (icp->type == ND_ROUTER_SOLICITATION) 
      recv_rtsol(ifp, ip, cc);
}
/*--END FUNCTION--(recv_icmp)-----------------------------------------------*/


/*--START FUNCTION--(recv_rtsol)--------------------------------------------

  FIXME: link-layer address option?
----------------------------------------------------------------------------*/
void
recv_rtsol(struct interface *ifp, register struct ipv6hdr *ip, int cc) {
   register struct icmpv6_hdr *icp;      /* ICMPv6 header */
   register struct nd_opt *hp;           /* ND option */
   /* struct prefix *pp;
      int changed = 0; */

#ifdef NDPD_ROUTER_DEBUG	
   printf("Received a recv_rtsol\n");
   dump_hex((char *)ip,  cc + sizeof(struct ipv6hdr));
#endif	
   
   icp = (struct icmpv6_hdr *)(ip + 1);

   /*
    * Check for anomalies
    */
   if (icp->code != 0) {
      if (verbose)
	 logerr("ICMPv6 Rt Sol: code=%d\n", icp->code);
      return;
   }
   
   if ((unsigned)cc < sizeof(struct r_sol)) {
      if (verbose)
	 logerr("ICMPv6 Rt Sol: too short %d\n", cc);
      return;
   }

   if (verbose) {
      logtrace("receive ICMPv6 Rt Sol on %s from %s",
	       ifp->name,
	       inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
			 INET6_ADDRSTRLEN));
      logtrace(" to %s (len=%d)\n",
	       inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
			 INET6_ADDRSTRLEN),
	       cc + sizeof(struct ipv6hdr));
   }

   cc -= sizeof(struct r_sol);
   
   /*
    * Read options
    */
   hp = (struct nd_opt *)((char *)icp + sizeof(struct r_sol));
   while (cc > 0) {
      switch (hp->type) {
	 /* 
	  * Kernel takes care of link-layer address option
	  * Only check option correctness
	  */
      default:
      next:
         if (hp->len == 0) {
	    if (verbose)
	       logerr("Null option length (type = %d)\n", hp->type);
	    return;
	 }
	 cc -= hp->len << 3;
	 
	 if (cc < 0) {
	    if (verbose)
	       logerr("Rt Sol underflow\n");
	    return;
	 }
	 if (cc == 0)
	    break;
	 hp = (struct nd_opt *)((char *)hp + (hp->len << 3));
      }
   }
   sol_rtadv(ifp);
}
/*--END FUNCTION--(recv_rtsol)----------------------------------------------*/


/*--START FUNCTION--(recv_rtadv)---------------------------------------------
  
  --------------------------------------------------------------------------*/
void
recv_rtadv(struct interface *ifp, register struct ipv6hdr *ip, int cc) {
   register struct r_adv *icp;           /* ICMPv6 header */
   register char *hp;                    /* Option header */
   struct prefix *pp;                    /* Prefix header */
   /* int changed = 0;
      __u8 tmp_managed; */
   struct in6_rtmsg rt6;                 /* Routing message to kernel */
	
   /*
    * Check errors
    */
   icp = (struct r_adv *)(ip + 1);
   if (icp->hdr.code != 0)
      return;
   if ((unsigned)cc < sizeof(struct r_adv))
      return;

   if (verbose) {
      logtrace("receive ICMPv6 Rt Adv on %s from %s\n",
	       ifp->name,
	       inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
			 INET6_ADDRSTRLEN));
   }
         
	
   if ((MaxHop && icp->hop_limit && MaxHop != icp->hop_limit) ||
       (Retrans && icp->retrans_t && Retrans != icp->retrans_t) ||
       ((Managed == 0x80) != icp->flags.managed) ||
       ((Managed == 0x40) != icp->flags.other) ||
       (Reach && icp->reachable_t && Reach != icp->reachable_t) ) {
      logerr("Rt Adv on %s from %s mismatch\n",
	     ifp->name,
	     inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
		       INET6_ADDRSTRLEN));
      logerr(
	     "hop %d!=%d or flags 0x%x!=0x%x or Retrans %d!=%d or RechTime %d!=%d\n",
	     MaxHop, icp->hop_limit,
	     Managed, icp->flags,
	     Retrans, icp->retrans_t,
	     Reach, icp->reachable_t);
   }

   cc -= sizeof(struct r_adv);
   
   /*
    * Options handling
    */
   (struct nd_opt *)hp = (struct nd_opt *)((char *)icp + sizeof(struct r_adv));
   while (cc) {
      /*
       * Check option length is not 0
       */
      if (((struct nd_opt *)hp)->len == 0) {
	 if (verbose)
	    logerr("null option length (%d)\n",
		   ((struct nd_opt *)hp)->type);
	 return;
      }
      
      switch (((struct nd_opt *)hp)->type) {
	 /*
	  * MTU option
	  */
      case ND_OPT_MTU:
	 if (MTU(hp)->mtu != ifp->mtu) {
	    logerr("Rt Adv on %s from %s mtu %d!=%d\n",
		   ifp->name,
		   inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
			     INET6_ADDRSTRLEN),
		   MTU(hp)->mtu,
		   ifp->mtu);
	 }
	 goto next;
	 
	 /*
	  * Prefix option
	  */
      case ND_OPT_PREFIX_INFO:
	 for (pp = ifp->prefixes.tqh_first;
	      pp && ((pp->plen != PREF(hp)->plen) || 
		     !SAME_ADDR6(pp->addr, PREF(hp)->prefix));
	      pp = pp->prf_list.tqe_next)
	    continue;
	 if (pp == NULL)
	    goto next;
	 
	 /*
	  * Lifetime changes?
	  */
	 if ((PREF(hp)->valid_lt != INFLIFE) ||
	     (PREF(hp)->pref_lt != INFLIFE) ) {
	    logerr("Rt Adv on %s from %s prefix mismatch\n",
		   ifp->name,
		   inet_ntop(AF_INET6, &ip->saddr, inet_ntop_ret,
			     INET6_ADDRSTRLEN));
	    logerr(
		   "for prefix %s/%d val_life %d!=%d or pref_life %d!=%d\n",
		   inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
			     INET6_ADDRSTRLEN),
		   PREF(hp)->valid_lt, INFLIFE,
		   PREF(hp)->pref_lt, INFLIFE);
	 }
	 goto next;
	 
      default:
      next:
	 cc -= ((struct nd_opt *)hp)->len << 3;
	 if (cc < 0)
	    return;

	 if (cc == 0)
	    break;
	 
	 (struct nd_opt *)hp = (struct nd_opt *)
	                       (hp + (((struct nd_opt *)hp)->len << 3));
      }
   }
   
   /* 
    * Save link local route to router
    */
   memset(&rt6, 0, sizeof(rt6));
   rt6.rtmsg_type      = SIOCADDRT;
   COPY_ADDR6(rt6.rtmsg_dst, ip->saddr);
   rt6.rtmsg_prefixlen = 128; /* local */
   strcpy(rt6.rtmsg_device, ifp->name);
   rt6.rtmsg_flags     = RTF_UP | RTF_HOST;	
	             
   if (ioctl(u, SIOCADDRT, &rt6) < 0) {
      if (verbose)
	 logerr("Socket error: Couldn't add route.");
      return;
   }
	
}
/*--END FUNCTION--(recv_rtadv)-----------------------------------------------*/














