#include "ndpd.h"

int ll_not_unique = 0;

/*--START FUNCTION--(recv_route)---------------------------------------------
  
  Handles messages received from kernel through netlink.
  
  --------------------------------------------------------------------------*/
void
recv_route()
{
   struct netlink_rtinfo kmsg;
   register struct sockaddr_in6 *sin, *sin_pref;
   register struct interface *ifp;
   struct prefix *pp;
   /*struct router *rp;
     struct timeout *tp;
     long delay;*/
   int cc;
   char *cp;
   u_char plen = 0;

   cp = (caddr_t) &kmsg;
   if (debug)
      bzero(cp, 2048);
   cc = read(r, cp, 2048);
   if (cc < 0) {
      if (verbose)
	 logperror("read (route message)");
      return;
   } else if (cc == 0) {
      if (verbose)
	 logerr("read (route message) return 0\n");
      return;
   }
       
   if (verbose)
      logtrace("receive route (%lu,%s,%d)\n",
	       kmsg.rtmsg_type, kmsg.rtmsg_device, kmsg.rtmsg_index);

   switch (kmsg.rtmsg_type) {
      /*
       * Address added to interface.
       * New prefix...
       */
   case RTMSG_NEWADDR:
      for (ifp = interfaces.tqh_first;
	   ifp && (ifp->index != kmsg.rtmsg_index);
	   ifp = ifp->if_list.tqe_next)
	 continue;
		
      if (ifp == NULL) {
	 if(verbose)
	    logerr("Don't take care of this interface\n");
	 return;
      }
		
      sin = &kmsg.rtmsg_dst6;
		 
      if (sin->sin6_family != AF_INET6) {
	 if(verbose)
	    logerr("Not AF_INET6\n");
	 return;
      }
		
      if ((sin == NULL) ||
	  IS_ANYADDR6(sin->sin6_addr) ||
	  IS_LINKLADDR6(sin->sin6_addr) ||
	  (sin->sin6_addr.s6_addr8[0] == 0) ||
	  (sin->sin6_addr.s6_addr8[0] == 0xff)) {
	 if(verbose)
	    logerr("Don't take care of this address type\n");
	 return;
      }
		
      sin_pref = &kmsg.rtmsg_gw6;
      plen = kmsg.rtmsg_plen6; 
      if (plen < 3) {
	 if(verbose)
	    logerr("Bad plength\n");
	 return;
      }
		
      if (verbose)
	 logtrace("get new address %s/%d on %s\n",
		  inet_ntop(AF_INET6, &sin->sin6_addr, 
			    inet_ntop_ret, INET6_ADDRSTRLEN),
		  plen, ifp->name);
      if (Calloc(pp, prefix) == NULL) {
	 if (verbose)
	    logerr("out of memory\n");
	 return;
      }

      pp->plen = plen;
      COPY_ADDR6(pp->addr, sin->sin6_addr);
      COPY_ADDR6(pp->pref, pp->addr);
      mk_pref(pp);
      mk_mask(&pp->amask, (int)plen);
      pp->ifp = ifp;
      pp->invlife = INFLIFE;
      pp->deplife = DEFDLIFE;
      pp->nflags = ND_PREF_FLG_A;
      if (pp->plen != 128)
	 pp->nflags |= ND_PREF_FLG_L;
      pp->pflags = pp->nflags;
      TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
      update_prefix(pp);
      sort_prefix(ifp);
      return;
		
      /*
       * Address removed from interface
       */
   case RTMSG_DELADDR:
      for (ifp = interfaces.tqh_first;
	   ifp && (ifp->index != kmsg.rtmsg_index);
	   ifp = ifp->if_list.tqe_next)
	 continue;
      if (ifp == NULL) {
	 if(verbose)
	    logerr("Don't take care of this interface\n");
	 return;
      }
		
      sin = &kmsg.rtmsg_dst6;
      if (sin->sin6_family != AF_INET6) {
	 if(verbose)
	    logerr("Not AF_INET6\n");
	 return;
      }
      if ((sin == NULL) ||
	  IS_ANYADDR6(sin->sin6_addr) ||
	  IS_LINKLADDR6(sin->sin6_addr) ||
	  (sin->sin6_addr.s6_addr8[0] == 0) ||
	  (sin->sin6_addr.s6_addr8[0] == 0xff)) {
	 if(verbose)
	    logerr("Don't take care of this address type\n");
	 return;
      }
		
      sin_pref = &kmsg.rtmsg_gw6;
      plen = kmsg.rtmsg_plen6;
      if (plen < 3) {
	 if(verbose)
	    logerr("Bad plength\n");
	 return;
      }
		
      for (pp = ifp->prefixes.tqh_first;
	   pp && ((pp->plen != plen) || 
		  !SAME_ADDR6(pp->addr, sin->sin6_addr));
	   pp = pp->prf_list.tqe_next)
	 continue;
      if (pp == NULL)
	 return;
      prefix_inval(pp);
      return;
		
      /*
       *  Change in interface status
       */
   case RTMSG_NEWDEVICE:
   case RTMSG_DELDEVICE:
      for (ifp = interfaces.tqh_first;
	   ifp && ((ifp->index != kmsg.rtmsg_index) ||
		   (((ifp->flags ^ kmsg.rtmsg_flags) & IFF_UP) == 0));
	   ifp = ifp->if_list.tqe_next)
	 continue;
      if (ifp == NULL)
	 return;

      ifp->flags = kmsg.rtmsg_flags;
      if ((ifp->flags & IFF_UP) == 0) {
	 if (verbose)
	    logtrace("interface %s down\n", ifp->name);
	 return;
      }

      if (verbose)
	 logtrace("interface %s up\n", ifp->name);

      start_rtadv(ifp);
      return;

   default:
      return;
   }
}

/*--START FUNCTION--(if_fetch)-----------------------------------------------
  
  
----------------------------------------------------------------------------*/
static int if_fetch(char *ifname, struct interface *ife)
{
   struct ifreq ifr;
 
   /* 
    * Name
    */
   strcpy(ife->name, ifname);

   /* 
    * Flags 
    */
   strcpy(ifr.ifr_name, ifname);
   if (ioctl(u, SIOCGIFFLAGS, &ifr) < 0) {
      fprintf(stderr, "Can't get interface %s 's flags.", ifname);
      return(-1);
   }
#define IFF_LAN (IFF_UP|IFF_BROADCAST|IFF_MULTICAST)
   if (((ifr.ifr_flags & IFF_LAN) != IFF_LAN)) {
      fprintf(stderr, "%s: Not a LAN interface\n", ifname);  
      return -1;
   }
   ife->flags = ifr.ifr_flags;
 
   /* 
    * Index 
    */
   if(ioctl(u, SIOGIFINDEX, &ifr) < 0){
      fprintf(stderr, "Can't get interface %s's index.", ifname);
      return (-1);
   } 
   ife->index = ifr.ifr_ifindex;

   /* 
    * Token (EUI 64)
    */
   strcpy(ifr.ifr_name, ifname);
   if (ioctl(u, SIOCGIFHWADDR, &ifr) < 0) {
      fprintf(stderr, "Can't get token for %s\n", ife->name);
      return -1; /* No token */
   }
   ife->token[0] = ifr.ifr_hwaddr.sa_data[0];
   ife->token[1] = ifr.ifr_hwaddr.sa_data[1];
   ife->token[2] = ifr.ifr_hwaddr.sa_data[2];
   ife->token[3] = 0xff;
   ife->token[4] = 0xfe;
   ife->token[5] = ifr.ifr_hwaddr.sa_data[3];
   ife->token[6] = ifr.ifr_hwaddr.sa_data[4];
   ife->token[7] = ifr.ifr_hwaddr.sa_data[5];

   /* 
    * MTU 
    */
   strcpy(ifr.ifr_name, ifname);
   if ((ioctl(u, SIOCGIFMTU, &ifr) < 0) || (ifr.ifr_mtu < IPV6_MTU)) {
      fprintf(stderr, "Can't get valid PMTU for %s\n", ife->name);
      return -1; /* Invalid MTU */
   }
   ife->mtu = ifr.ifr_mtu;
   return(0);
}


/*-------------------------------------------------------------------------
  Initializes daemon lists.
  
  ---------------------------------------------------------------------*/
void
init() {
   struct ifconf ifc;
   register struct ifreq *ifr;
   struct ifreq ifr2;
   
   register struct interface *ifp, *ifp2;
   register struct prefix *pp;
   struct sockaddr_in6 *sdl;
   int i, 
      itf_addr_nb;

   char *buf, *buf2;
   char buff[1024], buff2[1024];
	
   /*
    * Init lists
    */
   TAILQ_INIT(&interfaces);
   TAILQ_INIT(&timeouts);
   ifp = NULL;
 
   /*
    * Get all interfaces
    */
   bzero((caddr_t)&ifc, sizeof(ifc));
   bzero((caddr_t)&ifr, sizeof(ifr));
	
   ifc.ifc_len = sizeof(buff);
   ifc.ifc_buf = buff;
   ifr.ifr_addr.sa_family = AF_INET6;
   memcpy(buff, ((char *) &ifr), sizeof(struct ifreq));
   if (ioctl(u, SIOCGIFCONF, &ifc) < 0) {
      logperror("ioctl0");
      exit(4);
   }  
   
   /*
    * Scan interface list returned by kernel
    */
   buf = buff;
   i = ifc.ifc_len / sizeof(struct ifreq);
	
   for (; i > 0; i--,  buf+=sizeof(struct ifreq)){
      ifr = (struct ifreq *)buf;
      
      /*
       * New element in application interface list
       */
      if (Calloc(ifp, interface) == NULL) {
	 logerr(stderr, "out of memory\n");
	 abort();
      }
      memset((char *) ifp, 0, sizeof(struct interface));	   	
      TAILQ_INSERT_TAIL(&interfaces, ifp, if_list);
      TAILQ_INIT(&ifp->prefixes);      
     
      /*
       * Get interface parameters
       */
      if (if_fetch(ifr->ifr_name, ifp) < 0) {
	 fprintf(stderr, "%s: Will not handle this interface\n",
		 ifr->ifr_name);
	 continue;
      }
      
      /*bcopy(LLADDR(sdl), (caddr_t)ifp->token, LTOKEN);*/
      ifp->ntosend = 0;
      
      if (verbose)
	 logtrace("Read interface %s\n", ifp->name);
      
      /*
       * Retrieve all addresses 
       */
      ifr.ifr_addr.sa_family = AF_INET6;
      ifr.ifr_len6 = sizeof(buff2);
      ifr.ifr_buf6 = buff2;
      if (ioctl(u, SIOCGIFADDR, ifr) < 0) { 
	 fprintf(stderr, "Pb getting ipv6 addresses for %s\n", ifp->name);
	 return;
      }
      itf_addr_nb = ifr.ifr_len6 / sizeof(struct in6_addr);
      buf2 = buff2;
      
      /*
       * Update prefix list
       */
      for (; itf_addr_nb > 0; 
	   buf2 += sizeof(struct ifreq6), itf_addr_nb --) {
	 /* Alloc prefix struct memory */
	 if (Calloc(pp, prefix) == NULL) {
	    if (verbose)
	       logerr("Out of memory\n");
	    continue;
	 }
	 
	 /* Prefix length */
	 pp->plen = ((struct ifreq6 *)buf2)->preflen;
	 
	 /* IPv6 Address */
	 memcpy(&pp->addr, &(((struct ifreq6 *)buf2)->in6), 
		sizeof (struct in6_addr));
	 
	 if (verbose)
	    fprintf(stderr, "Interface %s has addr %s plen %d \n",
		    ifp->name,
		    inet_ntop(AF_INET6, &pp->addr, 
			      inet_ntop_ret, INET6_ADDRSTRLEN), 
		    pp->plen);
         
	 /* Null Addr??? */
	 if (IS_ANYADDR6(pp->addr)) 
	    goto loop_on_error;
	 
	 /* Link-local Addr? (Does not need be advertised!) */
	 if (IS_LINKLADDR6(pp->addr)) {
	    if (IS_ANYADDR6(ifp->llocal)) {
	       COPY_ADDR6(ifp->llocal, pp->addr);	
	       printf("Interface link-local update: % s\n", 
		      inet_ntop(AF_INET6, &pp->addr,
				inet_ntop_ret, INET6_ADDRSTRLEN));
	    }
	    goto loop_on_error;
	 }
	 
	 /* ??, multicast addr. */
	 if ((pp->addr.s6_addr8[0] == 0) ||
	     (pp->addr.s6_addr8[0] == 0xff))
	    goto loop_on_error;
	 
	 /* Interface */
	 pp->ifp        = ifp;
	 
	 /* Prefix */
	 COPY_ADDR6(pp->pref, pp->addr);	
	 mk_pref(pp);
	 
	 /* Life times */
	 pp->invlife    = INFLIFE;
	 pp->deplife    = DEFDLIFE;
	 
	 /* Flags */
	 pp->nflags     = ND_PREF_FLG_A;
	 if (pp->plen != 128)
	    pp->nflags |= ND_PREF_FLG_L;
	 pp->pflags     = pp->nflags;
	 
	 TAILQ_INSERT_TAIL(&ifp->prefixes, pp, prf_list);
      }
   loop_on_error:
      if (verbose)
	 logerr("Ignored interface address %s\n",
		inet_ntop(AF_INET6, &pp->addr,
			  inet_ntop_ret, INET6_ADDRSTRLEN));
      free(pp);
   } 

   ifp = interfaces.tqh_first;
   if (ifp)
      ifp = ifp->if_list.tqe_next;
   /* FIXME Does not look very useful to us */
   /*for (; ifp; ifp = ifp->if_list.tqe_next)
     if (SAME_ADDR6(interfaces.tqh_first->llocal, ifp->llocal)) {
     ll_not_unique = 1;
     break;
     }
     */
   for (ifp = interfaces.tqh_first; ifp;) {
      /*if (ll_not_unique && !FIRST(ifp)) {
	if (verbose)
	logtrace(
	"delete interface %s without uniq LL\n",
	ifp->name);
	goto del_if;
	}*/
      if (IS_ANYADDR6(ifp->llocal)) {
	 if (verbose)
	    logtrace("delete interface %s without LL\n",
		     ifp->name);
      del_if:
	 ifp2 = ifp;
	 ifp = ifp->if_list.tqe_next;
	 TAILQ_REMOVE(&interfaces, ifp2, if_list);
	 for (pp = FIRST(ifp2); pp; pp = pp->prf_list.tqe_next)
	    free(pp);
	 free(ifp2);
	 continue;
      }
      if (bcmp(&ifp->llocal.s6_addr8[16 - LTOKEN],
	       ifp->token, LTOKEN) != 0) {
	 if (verbose)
	    logtrace("update MAC of interface %s\n", ifp->name);
	 bcopy(&ifp->llocal.s6_addr8[16 - LTOKEN],
	       ifp->token, LTOKEN);
      }
      if (ll_not_unique) {
	 /*printf("ll_not_unique\n");*/
	 COPY_ADDR6(ifp->addr, FIRST(ifp)->addr);
      } else {
	 /*     printf("ll_unique\n");*/
	 COPY_ADDR6(ifp->addr, ifp->llocal);
      }
      for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next)
	 update_prefix(pp);
      sort_prefix(ifp);
      ifp = ifp->if_list.tqe_next;
   }
}
	
/*--END FUNCTION--(init)---------------------------------------------------*/

/*
int
is_router()
{
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IP6CTL_FORWARDING};
	int len = (sizeof mib)/(sizeof (int));
	int val;
	size_t size = sizeof(val);

	if (sysctl(mib, len, &val, &size, 0, 0) < 0 || size != sizeof(val)) {
		if (verbose)
			logerr("can't test ip6_forwarding\n");
		return (1);
	}
	return(val);
}
*/







