#include <string.h>
#include <stdlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include <sys/queue.h>

#include <net/if.h>
#include <net/dret/route6.h>
#include <net/dret/nd.h>
#include <net/netlink.h>

#include <linux/in.h>
#include <linux/dret/ipv6.h>
#include <linux/dret/icmpv6.h>
#include <linux/un.h>
#include <linux/route.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in_systm.h>

#include <asm/checksum.h>

#define INET6
/* #define NDPD_ROUTER_DEBUG */

#ifndef TAILQ_INSERT_BEFORE
#define	TAILQ_INSERT_BEFORE(listelm, elm, field) {			\
	(elm)->field.tqe_prev = (listelm)->field.tqe_prev;		\
	(elm)->field.tqe_next = (listelm);				\
	*(listelm)->field.tqe_prev = (elm);				\
	(listelm)->field.tqe_prev = &(elm)->field.tqe_next;		\
}
#endif

#define MAXTTL		255
#define	LTOKEN		8	/* EUI64 FIXME: should be variable ? */
#define WAIT_INIT	30

#define RTADV_FLG_M	0x80
#define RTADV_FLG_O	0x40
#define RTADV_FLG	(RTADV_FLG_M|RTADV_FLG_O)

#define MAX_RTR_SOLICITATIONS 		3
#define RTR_SOLICITATION_INTERVAL 	4.0
#define MAX_RTR_SOLICITATION_DELAY	1.0
/* from rfc1970
	MAX_INITIAL_RTR_ADVERT_INTERVAL  16 seconds
	MAX_INITIAL_RTR_ADVERTISEMENTS    3 transmissions
	MAX_FINAL_RTR_ADVERTISEMENTS      3 transmissions
	MIN_DELAY_BETWEEN_RAS             3 seconds
	MAX_RA_DELAY_TIME                 .5 seconds
	AdvSendAdvertisements	FALSE
	MaxRtrAdvInterval Default: 600 seconds
	MinRtrAdvInterval Default: 0.33 * MaxRtrAdvInterval
	AdvManagedFlag Default: FALSE
	AdvOtherConfigFlag Default: FALSE
	AdvLinkMTU Default: 0
	AdvReachableTime Default: 0
	AdvRetransTimer Default: 0
	AdvMaxHopLimit Default:  value in most recent "Assigned Numbers" RFC
	AdvDefaultLifetime Default: 3 * MaxRtrAdvInterval
	AdvPrefixList Each prefix has an associated:
	       AdvValidLifetime Default: infinity (0xffffffff)
	       AdvOnLinkFlag Default: TRUE
	       AdvPreferredLifetime Default: 604800 seconds (7 days)
*/
#define M_SECOND			1000	/* ms per sec */
#define MU_SECOND			1000	/* microsec per M_SECOND */
#define TO_SEC(x)			((x)/M_SECOND)
#define TO_USEC(x)			(((x)%M_SECOND)*MU_SECOND)
#define MAX_INITIAL_RTR_ADVERT_INTERVAL	(16*M_SECOND)
#define MAX_INITIAL_RTR_ADVERTISEMENTS	3
#define MAX_FINAL_RTR_ADVERTISEMENTS	3
#define MIN_DELAY_BETWEEN_RAS		(3*M_SECOND)
#define MAX_RA_DELAY_TIME		(M_SECOND/2)
#define DEF_RTR_ADV_INTERVAL		(600*M_SECOND)
#define MAX_RTR_ADV_INTERVAL		(1800*M_SECOND)
#define MIN_RTR_ADV_INTERVAL		(4*M_SECOND)
extern int MaxRtrAdvInterval, MinRtrAdvInterval;
extern u_int Retrans, Reach, MaxHop;
extern __u8 Managed;
        

#define Calloc(p, t)	(p = (struct t *)calloc(1,sizeof(struct t)))

struct prefix {
	TAILQ_ENTRY(prefix) prf_list;	/* prefix list */
	struct interface *ifp;		/* up pointer */
	struct in6_addr pref;		/* prefix itself */
	struct in6_addr pmask;		/* prefix mask */
	struct in6_addr addr;		/* address */
	struct in6_addr amask;		/* address mask */
	struct timeout *tp;		/* timeout */
	u_int invlife;			/* invalidation lifetime */
	u_int deplife;			/* deprecation lifetime */
	u_int rendelay;			/* renew delay */
	u_char plen;			/* length */
	u_char pflags;			/* flags */
	u_char nflags;			/* new flags */
	
};
#define INFLIFE		0xffffffff
#define DEFDLIFE	(7 * 24 * 60 * 60)

struct interface {
	TAILQ_ENTRY(interface) if_list;	/* list of interfaces */
	TAILQ_HEAD(, prefix) prefixes;	/* prefixes */
        unsigned int index;             /* interface index */
	struct in6_addr llocal;		/* link-local address */
	struct in6_addr addr;		/* best address */
	u_char token[LTOKEN];		/* MAC address or token */
	char autoconf;			/* autoconf enabled */
	u_char stateful;		/* stateful flags */
	u_long mtu;			/* IPv6 MTU */
	short flags;			/* copy of kernel flags */
	char name[IFNAMSIZ];		/* name */
	int ntosend;			/* Rt to send */
	struct timeout *tp;		/* when next */
	struct timeval last;		/* when last */
};
TAILQ_HEAD(, interface) interfaces;
struct interface *def_if, *main_if;
#define FIRST(ifp)	(ifp->prefixes.tqh_first)

struct timeout {
	TAILQ_ENTRY(timeout) tm_list;	/* timeout list */
	struct timeval tv;		/* when */
	void (*func) ();		/* function */
	void *arg;			/* argument */
	int expired;			/* expired */
	int refcount;			/* refcount */
};
TAILQ_HEAD(, timeout) timeouts;

struct {
	struct in6_addr	addr;		/* default source address */
	unsigned long  index;		/* default interface index */
} def_ifa;


extern int pid;

#define ROUNDUP(a) \
        ((a) > 0 ? (1 + (((a) - 1) | (sizeof(int32_t) - 1))) : sizeof(int32_t))
#define ADVANCE(x, n) (x += ROUNDUP(((struct sockaddr *)n)->sa_len))

extern struct timeval now;

#define NORMTIME(tv) \
	while ((tv).tv_usec >= 1000000) { \
		(tv).tv_sec += 1; \
		(tv).tv_usec -= 1000000; \
	}

#define AFTER(tv1, tv2) \
	(((tv1).tv_sec > (tv2).tv_sec) || \
	 (((tv1).tv_sec == (tv2).tv_sec) && ((tv1).tv_usec >= (tv2).tv_usec)))

extern int s, r, u;			/* Socket file descriptors */
extern char inet_ntop_ret[], inet_ntop_ret1[];

extern int routeradv;
extern int prefixadv;
extern int verbose;
extern int debug;
extern int timestamps;

void start_rtadv(struct interface *);
void stop_rtadv(struct interface *);
void sol_rtadv(struct interface *);
void set_router(struct interface *);
int in_cksum(u_short *, int);
void recv_icmp();
void recv_rtsol(struct interface *, struct ipv6hdr *, int);
void recv_rtadv(struct interface *, struct ipv6hdr *, int);
void recv_llredirect(struct interface *, struct ipv6hdr *, int);
void recv_rtredirect(struct interface *, struct ipv6hdr *, int);
void recv_route();
void init(), set_verbose(int), dump(), finish();
void get_main_if();
void update_prefix(struct prefix *);
void prefix_deprec(struct prefix *);
void prefix_inval(struct prefix *);
void sort_prefix(struct interface *);
void add_prefix(struct prefix *);
void del_prefix(struct prefix *);
void prompt(struct interface *);
void add_prefroute(struct prefix *);
void del_prefroute(struct prefix *);
void add_timeout(struct timeout *);
void dump_prefix(struct prefix *);
void dump_interface(struct interface *);
void dump_timeout(struct timeout *);
void mk_addr(struct prefix *);
void mk_mask(struct in6_addr *, int);
void mk_pref(struct prefix *);
u_char get_prefixlen(struct sockaddr_in6 *);
char *mac2ascii(caddr_t);
void initlog();
void logerr(), logtrace(), logdebug();
void logperror(char *);
int is_router();








