#include "ndpd.h"

struct router *def_rt;
int booting = 1;

/*
 * Update prefix information
 */
void
update_prefix(register struct prefix *pp) {
   (void)gettimeofday(&now, NULL);

   /*
    * Check timeout list
    */
   if (pp->tp) {
      if (pp->tp->expired == 0)
	 TAILQ_REMOVE(&timeouts, pp->tp, tm_list);
      if (--pp->tp->refcount <= 0)
	 free(pp->tp);
      pp->tp = NULL;
   }

   /*
    * Update prefix flags
    */
   pp->nflags &= ~1;
   if ((pp->pflags != pp->nflags) && (pp->nflags & ND_PREF_FLG_A)) {
      CLR_ADDR6(pp->amask);
      if (pp->nflags & ND_PREF_FLG_L) {
	 COPY_ADDR6(pp->pmask, pp->amask);
      } else {
	 mk_mask(&pp->amask, 128);
      }
   }
   pp->pflags = pp->nflags;

   /*
    * Check lifetimes
    */
   if (pp->invlife == 0) {
      prefix_inval(pp);
      return;
   }
   
   if (pp->deplife == 0) {
      prefix_deprec(pp);
      return;
   }

   /*
    * New timeouts
    */
   if (Calloc(pp->tp, timeout) == NULL) {
      if (verbose)
	 logerr("out of memory\n");
      return;
   }
		
   if (pp->deplife != INFLIFE) {
      pp->tp->tv.tv_sec = now.tv_sec + pp->deplife;
      pp->tp->tv.tv_usec = now.tv_usec;
      pp->tp->func = prefix_deprec;
      pp->tp->arg = pp;
      pp->tp->refcount++;
      add_timeout(pp->tp);
   } else if (pp->invlife != INFLIFE) {
      pp->tp->tv.tv_sec = now.tv_sec + pp->invlife;
      pp->tp->tv.tv_usec = now.tv_usec;
      pp->tp->func = prefix_inval;
      pp->tp->arg = pp;
      pp->tp->refcount++;
      add_timeout(pp->tp);
   } else {
      free(pp->tp);
      pp->tp = NULL;
   }
}

void
prefix_deprec(pp)
	register struct prefix *pp;
{
	struct prefix *first = FIRST(pp->ifp);

	(void)gettimeofday(&now, NULL);

	if (pp->tp) {
		if (pp->tp->expired == 0)
			TAILQ_REMOVE(&timeouts, pp->tp, tm_list);
		if (--pp->tp->refcount <= 0)
			free(pp->tp);
		pp->tp = NULL;
	}

	TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);
	TAILQ_INSERT_TAIL(&pp->ifp->prefixes, pp, prf_list);

	if (pp->invlife != INFLIFE) {
		if (Calloc(pp->tp, timeout) == NULL) {
			if (verbose)
				logerr("out of memory\n");
			return;
		}

		pp->tp->tv.tv_sec = now.tv_sec + pp->invlife;
		pp->tp->tv.tv_usec = now.tv_usec;
		pp->tp->func = prefix_inval;
		pp->tp->arg = pp;
		pp->tp->refcount++;
		add_timeout(pp->tp);
	}

	if (pp == first) {
		if (verbose)
			logtrace("keep deprecated %s on %s\n",
				 inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
					    sizeof(struct in6_addr)),
				 pp->ifp->name);
		sort_prefix(pp->ifp);
	}
}

void
prefix_inval(pp)
	register struct prefix *pp;
{
	struct prefix *first = FIRST(pp->ifp);

	if (pp->tp) {
		if (pp->tp->expired == 0)
			TAILQ_REMOVE(&timeouts, pp->tp, tm_list);
		if (--pp->tp->refcount <= 0)
			free(pp->tp);
		pp->tp = NULL;
	}

	TAILQ_REMOVE(&pp->ifp->prefixes, pp, prf_list);

	if (pp == first)
		sort_prefix(pp->ifp);
	free(pp);
}

/*
 * Sort prefixes
 */
void
sort_prefix(register struct interface *ifp) {
   register struct prefix *pp, *pp2;
   int changed = 0;

   /*
    * No prefix attached to interface
    */
   if (FIRST(ifp) == NULL) {
      COPY_ADDR6(ifp->addr, ifp->llocal);
      changed++;
   }

#define SWAP_PREFIX	\
   if (pp == FIRST(ifp)) \
      changed++; \
   TAILQ_REMOVE(&ifp->prefixes, pp2, prf_list); \
   TAILQ_INSERT_BEFORE(pp, pp2, prf_list); \
   if (pp == FIRST(ifp)) \
      changed++; \
   goto again;

again: 
   /*
    * Sort prefixes by lifetime and scope
    */
   for (pp = FIRST(ifp); pp; pp = pp2) {
      pp2 = pp->prf_list.tqe_next;
      if (pp2 == NULL)
	 break;
      if (IS_SITELADDR6(pp->addr)) {
	 if (!IS_SITELADDR6(pp2->addr)) {
	    SWAP_PREFIX
	 }
	 if ((pp->deplife < pp2->deplife) ||
	     ((pp->deplife == pp2->deplife) &&
	      (pp->invlife < pp2->invlife))) {
	    SWAP_PREFIX
	 }
      } else if (IS_SITELADDR6(pp2->addr)) {
	 continue;
      } else if ((pp->deplife < pp2->deplife) ||
		 ((pp->deplife == pp2->deplife) &&
		  (pp->invlife < pp2->invlife))) {
	 SWAP_PREFIX
      }
   }
}

