#include "ndpd.h"

void
add_timeout(register struct timeout *tp) {
   register struct timeout *tp2;

   /*
    * Timeout list is empty...
    */
   if (timeouts.tqh_first == NULL) {
      TAILQ_INSERT_HEAD(&timeouts, tp, tm_list);
      return;
   }

   /*
    * Find chronological location in list
    */
   for (tp2 = timeouts.tqh_first;
	tp2 && AFTER(tp->tv, tp2->tv);
	tp2 = tp2->tm_list.tqe_next)
      continue;

   /*
    * Tail insertion
    */
   if (tp2 == NULL) {
      TAILQ_INSERT_TAIL(&timeouts, tp, tm_list);
      return;
   }

   /*
    * Middle list insertion
    */
   TAILQ_INSERT_BEFORE(tp2, tp, tm_list);
}

void
dump_prefix(pp)
	register struct prefix *pp;
{
	logdebug("prefix (%lx):\n", (u_long)pp);
	logdebug("\tinterface=%s\n", pp->ifp->name);
	logdebug("\tprefix=%s ",
		 inet_ntop(AF_INET6, &pp->pref, inet_ntop_ret,
			    INET6_ADDRSTRLEN));
	logdebug("with len=%d and mask=%s\n",
		 pp->plen,
		 inet_ntop(AF_INET6, &pp->pmask, inet_ntop_ret,
			    INET6_ADDRSTRLEN));
	logdebug("\tif address=%s ",
		 inet_ntop(AF_INET6, &pp->addr, inet_ntop_ret,
			    INET6_ADDRSTRLEN));
	logdebug("with mask=%s\n",
		 inet_ntop(AF_INET6, &pp->amask, inet_ntop_ret,
			    INET6_ADDRSTRLEN));
	logdebug("\ttimeout=%lx\n", (u_long)(pp->tp));
	logdebug("\trenew=%u\n", pp->rendelay);
	if (pp->deplife == INFLIFE)
		logdebug("\tdeprecation=infinity\n");
	else
		logdebug("\tdeprecation=%u\n", pp->deplife);
	if ((pp->deplife == INFLIFE) || (pp->invlife == INFLIFE))
		logdebug("\tinvalidation=infinity\n");
	else
		logdebug("\tinvalidation=%u\n", pp->invlife);
	logdebug("\tflags=%x (next=%x)\n", pp->pflags, pp->nflags);
}

void
dump_interface(ifp)
	register struct interface *ifp;
{
	register struct prefix *pp;

	logdebug("interface (%lx):\n", (u_long)ifp);
	logdebug("\tname=%s\n", ifp->name);
	logdebug("\tindex=%d\n", ifp->index);
	logdebug("\tlink-local address=%s\n",
		 inet_ntop(AF_INET6, &ifp->llocal, inet_ntop_ret,
			   INET6_ADDRSTRLEN));
	logdebug("\tMTU=%lu\n", ifp->mtu);
	logdebug("\tflags=%x\n", ifp->flags);

	logdebug("prefix list:\n");
	for (pp = FIRST(ifp); pp; pp = pp->prf_list.tqe_next)
		dump_prefix(pp);
}

void
dump_timeout(tp)
	register struct timeout *tp;
{
	struct timeval delta;
	int sign = 0;

	logdebug("timeout (%lx):\n", (u_long)tp);
	if (AFTER(tp->tv, now)) {
		sign = +1;
		delta.tv_sec = tp->tv.tv_sec - now.tv_sec - 1;
		delta.tv_usec = tp->tv.tv_usec + 1000000 - now.tv_usec;
		NORMTIME(delta);
	} else {
		sign = -1;
		delta.tv_sec =  now.tv_sec - tp->tv.tv_sec - 1;
		delta.tv_usec = now.tv_usec + 1000000 - tp->tv.tv_usec;
		NORMTIME(delta);
	}
	logdebug("\texpire=%+ld.%06lu\n", sign * delta.tv_sec, delta.tv_usec);
	logdebug("\texpired=%d\n", tp->expired);
	logdebug("\trefcount=%d\n", tp->refcount);
	logdebug("\tfunction=%lx\n", (u_long)(tp->func));
	logdebug("\targument=%lx\n", (u_long)(tp->arg));
}

/*
 * Build address mask from prefix length
 */
void
mk_mask(register struct in6_addr *addr, register int len) {
   register u_char *cp;

   CLR_ADDR6(*addr);
   cp = (u_char *)addr;
   for (cp = (u_char *)addr; len > 7; len -= 8)
      *cp++ = 0xff;
   if(len)
      *cp = 0xff << (8 - len);
}

/*
 * Build prefix from addr and prefix length
 */
void
mk_pref(register struct prefix *pp) {
   register int i;

   mk_mask(&pp->pmask, pp->plen);
   for (i = 0; i < 16; i++)
      pp->pref.s6_addr8[i] &= pp->pmask.s6_addr8[i];
}

u_char
get_prefixlen(sdl)
	struct sockaddr_in6 *sdl;
{
	register u_char *cp;
	register int byte, bit, len;

	cp = (u_char *)sdl;
	for (byte = 2, cp += 2; byte < 8; byte++, cp++)
		if (*cp)
			return(0);

	len = 0;
	for (byte = 0; byte < 16; byte++, cp++, len += 8)
		if (*cp != 0xff)
			break;
	if (len == 128)
		return((u_char)128);

	for (bit = 7; bit > 0; bit--, len++)
		if (!(*cp & (1 << bit)))
			break;
	for (; bit > 0; bit--)
		if (*cp & (1 << bit))
			return(0);

	byte++; cp++;
	for (; byte < 16; byte++, cp++)
		if (*cp)
			return(0);
	return((u_char)len);
}

#if (LTOKEN == 6)
char *
mac2ascii(addr)
	register caddr_t addr;
{
	static char buf[64];

	sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x",
		(u_char) addr[0], (u_char) addr[1], (u_char) addr[2],
		(u_char) addr[3], (u_char) addr[4], (u_char) addr[5]);
	return(buf);
}
#endif

static logging = 0;

void
initlog() {
   logging++;
   openlog("ndpd-router", LOG_PID | LOG_CONS, LOG_DAEMON);
}

void
logdate()
{
	time_t now;
	char *timestamp;

	if (!timestamps)
		return;
	(void)time(&now);
	timestamp = ctime(&now);
	timestamp[19] = '\0';
	(void) fprintf(stderr, "%s: ", timestamp + 11);
}

/* VARARGS1 */
void
logerr(fmt, a,b,c,d,e,f,g,h)
	char *fmt;
{
	if (logging)
		syslog(LOG_ERR, fmt, a,b,c,d,e,f,g,h);
	else {
		logdate();
		(void) fprintf(stderr, fmt, a,b,c,d,e,f,g,h);
	}
}

/* VARARGS1 */
void
logtrace(fmt, a,b,c,d,e,f,g,h)
	char *fmt;
{
	if (logging)
		syslog(LOG_INFO, fmt, a,b,c,d,e,f,g,h);
	else {
		logdate();
		(void) fprintf(stdout, fmt, a,b,c,d,e,f,g,h);
	}
}

/* VARARGS1 */
void
logdebug(fmt, a,b,c,d,e,f,g,h)
	char *fmt;
{
	if (logging)
		syslog(LOG_DEBUG, fmt, a,b,c,d,e,f,g,h);
	else {
		logdate();
		(void) fprintf(stdout, fmt, a,b,c,d,e,f,g,h);
	}
}

void
logperror(str)
	char *str;
{
	if (logging)
		syslog(LOG_ERR, "%s: %s\n", str, sys_errlist[errno]);
	else {
		logdate();
		(void) fprintf(stderr, "%s: %s\n", str, sys_errlist[errno]);
	}
}

